# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""atomic quantizer implementation."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc
import math

import numpy as np
from qkeras import quantizers

FLOATINGPOINT_BITS = 32


def get_exp(quantizer):
  """get max/min exp value for relu_po2 or quantized_po2."""

  if quantizer.is_signed:
    non_sign_bits = quantizer.bits - 1
  else:
    non_sign_bits = quantizer.bits

  min_exp = -2 ** (non_sign_bits - 1)
  max_exp_orig = 2 ** (non_sign_bits - 1) - 1

  max_exp = max_exp_orig
  # max_value caps how many int_bits actually allowed
  if quantizer.max_val_po2 != -1:
    if quantizer.max_val_po2 <= 0:
      max_exp = 0
    else:
      max_exp = math.ceil(np.log2(quantizer.max_val_po2))
      max_exp = min(max_exp, max_exp_orig)

  # if max_exp<0. no need to expand int_bits
  max_exp = max(0, max_exp)

  return (-min_exp, max_exp)


class IQuantizer(abc.ABC):
  """abstract class for quantizer."""

  def __init__(self):
    self.mode = -1
    self.bits = -1
    self.int_bits = -1
    self.is_signed = 0
    self.is_floating_point = False
    self.max_val_po2 = -1
    self.is_po2 = 0
    self.name = None
    self.op_type = "quantizer"


class QuantizedBits(IQuantizer):
  """quantized bits.

  Attributes:
    mode: index of the current quantizer in
          MultiplierFactory.multiplier_impl_table
    bits: total bits
    int_bits: integer bits
    is_signed: if a signed number
    name: quantizer name
  """

  def __init__(self):
    super().__init__()
    self.mode = 0
    self.is_signed = 1
    self.name = "quantized_bits"

  def convert_qkeras_quantizer(
      self, quantizer: quantizers.quantized_bits):
    self.mode = 0
    self.bits = quantizer.bits
    self.int_bits = quantizer.integer
    self.is_signed = quantizer.keep_negative

  def convert_to_qkeras_quantizer(
      self, symmetric=1, alpha=None, use_stochastic_rounding=False,
      scale_axis=None, qnoise_factor=1.0):
    """convert qtools quantizer to qkeras quantizer."""

    return quantizers.quantized_bits(
        bits=self.bits, integer=self.int_bits, keep_negative=self.is_signed,
        symmetric=symmetric, alpha=alpha,
        use_stochastic_rounding=use_stochastic_rounding,
        scale_axis=scale_axis, qnoise_factor=qnoise_factor)


class QuantizedTanh(QuantizedBits):
  """same as quantized bits."""

  def __init__(self):
    super().__init__()
    self.name = "quantized_tanh"

  def convert_qkeras_quantizer(
      self, quantizer: quantizers.quantized_tanh):
    self.mode = 0
    self.bits = quantizer.bits
    self.is_signed = 1

  def convert_to_qkeras_quantizer(
      self, symmetric=False, use_stochastic_rounding=False):
    """convert qtools quantizer to qkeras quantizer."""

    return quantizers.quantized_tanh(
        bits=self.bits, use_stochastic_rounding=use_stochastic_rounding,
        symmetric=symmetric)


class QuantizedUlaw(QuantizedBits):
  """quantized ulaw type."""

  # same as quantized bits
  def __init__(self):
    super().__init__()
    self.name = "quantized_ulaw"

  def convert_qkeras_quantizer(
      self, quantizer: quantizers.quantized_ulaw):
    self.mode = 0
    self.bits = quantizer.bits
    self.int_bits = quantizer.integer
    self.is_signed = 1

  def convert_to_qkeras_quantizer(self, symmetric=0, u=255.0):
    """convert qtools quantizer to qkeras quantizer."""

    return quantizers.quantized_ulaw(
        bits=self.bits, integer=self.int_bits, symmetric=symmetric, u=u)


class Binary(IQuantizer):
  """binary quantizer."""

  def __init__(self, use_01=False):
    super().__init__()
    if use_01:
      self.mode = 4
      self.is_signed = 0
    else:
      self.mode = 3
      self.is_signed = 1

    self.bits = 1
    self.int_bits = 1
    self.use_01 = use_01
    self.name = "binary"

  def convert_qkeras_quantizer(self, quantizer: quantizers.binary):
    if quantizer.use_01:
      self.mode = 4
      self.is_signed = 0
    else:
      self.mode = 3
      self.is_signed = 1

    self.use_01 = quantizer.use_01

  def convert_to_qkeras_quantizer(self, alpha=None,
                                  use_stochastic_rounding=False):
    """convert qtools quantizer to qkeras quantizer."""

    return quantizers.binary(use_01=self.use_01, alpha=alpha,
                             use_stochastic_rounding=use_stochastic_rounding)


class StochasticBinary(Binary):
  """stochastic binary quantizer."""

  # same as binary(-1, 1)
  def __init__(self):
    super().__init__(use_01=False)
    self.name = "stochastic_binary"

  def convert_qkeras_quantizer(
      self, quantizer: quantizers.stochastic_binary):
    """convert qkeras quantizer to qtools quantizer."""

    pass

  def convert_to_qkeras_quantizer(self, alpha=None, temperature=6.0,
                                  use_real_sigmoid=True):
    """convert qtools quantizer to qkeras quantizer."""

    return quantizers.stochastic_binary(alpha=alpha, temperature=temperature,
                                        use_real_sigmoid=use_real_sigmoid)


class Bernoulli(Binary):
  """bernoulli quantizer. same as binary(0, 1)."""

  def __init__(self):
    super().__init__(use_01=True)
    self.name = "bernoulli"

  def convert_qkeras_quantizer(self, quantizer: quantizers.bernoulli):
    pass

  def convert_to_qkeras_quantizer(self, alpha=None, temperature=6.0,
                                  use_real_sigmoid=True):
    """convert qtools quantizer to qkeras quantizer."""

    return quantizers.bernoulli(alpha=alpha, temperature=temperature,
                                use_real_sigmoid=use_real_sigmoid)


class QuantizedRelu(IQuantizer):
  """quantized relu quantizer."""

  def __init__(self):
    super().__init__()
    self.is_signed = 0
    self.name = "quantized_relu"

  def convert_qkeras_quantizer(
      self, quantizer: quantizers.quantized_relu):
    """convert from qkeras quantizer."""

    bits = quantizer.bits
    int_bits = quantizer.integer

    if bits == 1 and int_bits == 1:
      mode = 4
    else:
      mode = 0

    self.mode = mode
    self.bits = bits
    self.int_bits = int_bits
    if hasattr(quantizer, "negative_slope") and quantizer.negative_slope != 0:
      self.is_signed = 1

  def convert_to_qkeras_quantizer(
      self, use_sigmoid=0, negative_slope=0.0, use_stochastic_rounding=False,
      relu_upper_bound=None, is_quantized_clip=True, qnoise_factor=1.0):
    """convert qtools quantizer to qkeras quantizer."""

    return quantizers.quantized_relu(
        bits=self.bits, integer=self.int_bits, use_sigmoid=use_sigmoid,
        negative_slope=negative_slope,
        use_stochastic_rounding=use_stochastic_rounding,
        relu_upper_bound=relu_upper_bound,
        is_quantized_clip=is_quantized_clip,
        qnoise_factor=qnoise_factor)


class Ternary(IQuantizer):
  """ternary(0, 1, -1)."""

  def __init__(self):
    super().__init__()
    self.mode = 2
    self.bits = 2
    self.int_bits = 2
    self.is_signed = 1
    self.name = "ternary"

  def convert_qkeras_quantizer(
      self, quantizer: quantizers.ternary):
    pass

  def convert_to_qkeras_quantizer(
      self, alpha=None, threshold=None, use_stochastic_rounding=False,
      number_of_unrolls=5):
    """convert qtools quantizer to qkeras quantizer."""

    return quantizers.ternary(
        alpha=alpha, threshold=threshold,
        use_stochastic_rounding=use_stochastic_rounding,
        number_of_unrolls=number_of_unrolls)


class StochasticTernary(Ternary):
  """stochastic ternary."""

  def __init__(self):
    super().__init__()
    self.name = "stochastic_ternary"

  # same as ternary
  def convert_qkeras_quantizer(
      self, quantizer: quantizers.stochastic_ternary):
    pass

  def convert_to_qkeras_quantizer(
      self, alpha=None, threshold=None, temperature=8.0,
      use_real_sigmoid=True, number_of_unrolls=5):
    """convert qtools quantizer to qkeras quantizer."""

    return quantizers.stochastic_ternary(
        alpha=alpha, threshold=threshold, temperature=temperature,
        use_real_sigmoid=use_real_sigmoid,
        number_of_unrolls=number_of_unrolls)


class FloatingPoint(IQuantizer):
  """float32."""

  def __init__(self, bits):
    super().__init__()
    self.mode = 5
    self.bits = bits
    self.int_bits = -1
    self.is_signed = 1
    self.is_floating_point = True
    self.name = "floating_point"

  def convert_qkeras_quantizer(self, bits):
    pass

  def convert_to_qkeras_quantizer(self, bits):
    pass


class PowerOfTwo(IQuantizer):
  """po2."""

  def __init__(self, is_signed=True):
    super().__init__()
    self.mode = 1
    self.is_po2 = 1
    self.is_signed = is_signed
    self.inference_value_counts = -1

    if is_signed:
      self.name = "quantized_po2"
    else:
      self.name = "quantized_relu_po2"

  def convert_qkeras_quantizer(self, quantizer):
    """convert qkeras quantizer to qtools quantizer."""

    assert "po2" in quantizer.__class__.__name__

    if quantizer.__class__.__name__ == "quantized_po2":
      self.is_signed = 1
      self.name = "quantized_po2"

    elif quantizer.__class__.__name__ == "quantized_relu_po2":
      super().__init__()
      self.is_signed = 0
      self.name = "quantized_relu_po2"

    bits = quantizer.bits
    max_val_po2 = quantizer.max_value
    if not max_val_po2:
      self.max_val_po2 = -1
    else:
      self.max_val_po2 = max_val_po2
    self.bits = bits
    self.int_bits = bits

  def convert_to_qkeras_quantizer(
      self, negative_slope=0, use_stochastic_rounding=False,
      quadratic_approximation=False):
    """convert qtools quantizer to qkeras quantizer."""

    if self.is_signed:
      # quantized_po2
      return quantizers.quantized_po2(
          bits=self.bits,
          max_value=self.max_val_po2 if self.max_val_po2 >= 0 else None,
          use_stochastic_rounding=use_stochastic_rounding,
          quadratic_approximation=quadratic_approximation)
    else:
      # quantized_relu_po2
      return quantizers.quantized_relu_po2(
          bits=self.bits,
          max_value=self.max_val_po2 if self.max_val_po2 >= 0 else None,
          negative_slope=negative_slope,
          use_stochastic_rounding=use_stochastic_rounding,
          quadratic_approximation=quadratic_approximation)

  def get_min_max_exp(self):
    return get_exp(self)

  def quantizer_bits_calculator(self, val):
    """calculate how many bits needed."""

    # calculate how many bits are required to represent a po2 value.
    # val can be +/- values, can be integer or franctional number.
    # needs to be dealt seperately.

    sign_bit = val < 0

    # get rid of sign
    val = abs(val)

    if val == 0:
      # val of 0 is special case; qkeras uses mininmum
      # number to represent 0
      non_sign_bits = self.bits - sign_bit
    else:
      exp_value = np.log2(val)

      # exp_value should be integer
      if abs(np.round(exp_value) - exp_value) > 0:
        raise ValueError("ERROR: {} is not a po2 value!".format(val))

      exp_value = int(exp_value)

      # for n bits, the range of values it can represent is:
      # min_val = -2 ** (n - 1)
      # max_val = 2 ** (n - 1) - 1
      if exp_value == 0:
        non_sign_bits = 1
      elif exp_value > 0:
        # e.g., 16 needs 5 bits + 1 exp sign bit,
        # 15 needs 4 bits + 1 exp sign bit
        non_sign_bits = math.floor(np.log2(exp_value)) + 1 + 1
      else:
        # e.g., -16 needs 4 bits + 1 exp sign bit
        non_sign_bits = math.ceil(np.log2(abs(exp_value))) + 1

    return (sign_bit, non_sign_bits)

  def update_quantizer(self, val, reset=False):
    """update quantizer bits according to the input value.

    Args:
      val: input value
      reset: True->disregard current quantizer bits and reset
        it according to the given value; False-> update the quantizer
        bits with given value.
        quantizer.bits = min(existing_bits, bits required by val)

    Returns:
      Update existing po2 quantizer bits by val.
       quantizer.bits = min(existing_bits, bits required by val)
    """
    (sign_bit, non_sign_bits) = self.quantizer_bits_calculator(val)

    if reset:
      self.bits = sign_bit + non_sign_bits
    else:
      # avoid input value exceeding quantizer limit
      self.bits = min(self.bits, sign_bit + non_sign_bits)

    self.int_bits = self.bits
    self.max_val_po2 = min(val, self.max_val_po2)
    self.is_signed = sign_bit

    if sign_bit:
      self.name = "quantized_po2"
    else:
      self.name = "quantized_relu_po2"

  def update_inference_values(self, weights):
    """find how many different values in weights in the po2 quantizer."""

    inference_value_counts = len(set(weights.flatten()))
    self.inference_value_counts = inference_value_counts


class ReluPowerOfTwo(PowerOfTwo):
  """relu po2."""

  def __init__(self):
    super().__init__()
    self.mode = 1
    self.is_po2 = 1
    self.is_signed = 0
    self.name = "quantized_relu_po2"

  def convert_qkeras_quantizer(
      self, quantizer: quantizers.quantized_relu_po2):

    self.bits = quantizer.bits
    self.int_bits = quantizer.bits
    if not quantizer.max_value:
      self.max_val_po2 = -1
    else:
      self.max_val_po2 = quantizer.max_value

  def convert_to_qkeras_quantizer(
      self, negative_slope=0, use_stochastic_rounding=False,
      quadratic_approximation=False):
    """convert qtools quantizer to qkeras quantizer."""

    # quantized_relu_po2
    return quantizers.quantized_relu_po2(
        bits=self.bits,
        max_value=self.max_val_po2 if self.max_val_po2 >= 0 else None,
        negative_slope=negative_slope,
        use_stochastic_rounding=use_stochastic_rounding,
        quadratic_approximation=quadratic_approximation)



# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
""""create divider quantizer."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc

from absl import logging
from qkeras.qtools.quantized_operators import divider_impl
from qkeras.qtools.quantized_operators import quantizer_impl


class UnacceptedQuantizerError(ValueError):
  pass


class IDivider(abc.ABC):
  """abstract class for divider."""

  def __init__(self):
    # also attached the output datatype in the table
    self.divider_impl_table = [
        [
            # when qbits is denominator, use default bits for float result
            (divider_impl.FloatingPointDivider, quantizer_impl.FloatingPoint(
                bits=quantizer_impl.FLOATINGPOINT_BITS)),
            (divider_impl.Shifter, quantizer_impl.QuantizedBits()),
            (None, None),
            (None, None),
            (None, None),
            # when bits sets to None, will decide f16/f32 according
            # to input quantizer
            (divider_impl.FloatingPointDivider, quantizer_impl.FloatingPoint(
                bits=None))
        ],
        [
            (divider_impl.FloatingPointDivider, quantizer_impl.FloatingPoint(
                bits=quantizer_impl.FLOATINGPOINT_BITS)),
            (divider_impl.Subtractor, quantizer_impl.PowerOfTwo()),
            (None, None),
            (None, None),
            (None, None),
            (divider_impl.FloatingPointDivider, quantizer_impl.FloatingPoint(
                bits=None))
        ],
        [
            (divider_impl.FloatingPointDivider, quantizer_impl.FloatingPoint(
                bits=quantizer_impl.FLOATINGPOINT_BITS)),
            (divider_impl.Shifter, quantizer_impl.QuantizedBits()),
            (None, None),
            (None, None),
            (None, None),
            (divider_impl.FloatingPointDivider, quantizer_impl.FloatingPoint(
                bits=None))
        ],
        [
            (divider_impl.FloatingPointDivider, quantizer_impl.FloatingPoint(
                bits=quantizer_impl.FLOATINGPOINT_BITS)),
            (divider_impl.Shifter, quantizer_impl.PowerOfTwo()),
            (None, None),
            (None, None),
            (None, None),
            (divider_impl.FloatingPointDivider, quantizer_impl.FloatingPoint(
                bits=None))
        ],
        [
            (divider_impl.FloatingPointDivider, quantizer_impl.FloatingPoint(
                bits=quantizer_impl.FLOATINGPOINT_BITS)),
            (divider_impl.Shifter, quantizer_impl.PowerOfTwo()),
            (None, None),
            (None, None),
            (None, None),
            (divider_impl.FloatingPointDivider, quantizer_impl.FloatingPoint(
                bits=None))
        ],
        [
            (divider_impl.FloatingPointDivider, quantizer_impl.FloatingPoint(
                bits=None)),
            (divider_impl.FloatingPointDivider, quantizer_impl.FloatingPoint(
                bits=None)),
            (None, None),
            (None, None),
            (None, None),
            (divider_impl.FloatingPointDivider, quantizer_impl.FloatingPoint(
                bits=None))
        ]
    ]

  def make_quantizer(self, numerator_quantizer: quantizer_impl.IQuantizer,
                     denominator_quantizer: quantizer_impl.IQuantizer):
    """make the quantizer."""

    self.numerator_quantizer = numerator_quantizer
    self.denominator_quantizer = denominator_quantizer

    mode1 = numerator_quantizer.mode
    mode2 = denominator_quantizer.mode

    (divider_impl_class, output_quantizer) = self.divider_impl_table[
        mode1][mode2]

    if divider_impl_class is None:
      raise UnacceptedQuantizerError(
          "denominator quantizer {} not accepted!".format(
              denominator_quantizer.name))

    logging.debug(
        "qbn adder implemented as class %s",
        divider_impl_class.implemented_as())

    return divider_impl_class(
        numerator_quantizer,
        denominator_quantizer,
        output_quantizer
    )



# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import warnings
import tensorflow as tf
from tensorflow.keras import constraints
from tensorflow.keras import initializers
from tensorflow.keras import regularizers
from tensorflow.keras.layers import Activation
from tensorflow.keras.layers import Conv1D
from tensorflow.keras.layers import Conv2D
from tensorflow.keras.layers import Conv2DTranspose
from tensorflow.keras.layers import SeparableConv1D
from tensorflow.keras.layers import SeparableConv2D
from tensorflow.keras.layers import DepthwiseConv2D
from tensorflow.keras.layers import Dropout
from tensorflow.keras.layers import InputSpec
from tensorflow.python.eager import context
from tensorflow.python.keras.utils import conv_utils
from tensorflow.python.ops import array_ops
from .qlayers import get_auto_range_constraint_initializer
from .qlayers import QActivation
from .quantizers import get_quantized_initializer
from .quantizers import get_quantizer

from tensorflow_model_optimization.python.core.sparsity.keras.prunable_layer import PrunableLayer


class QConv1D(Conv1D, PrunableLayer):
  """1D convolution layer (e.g. spatial convolution over images)."""

  # most of these parameters follow the implementation of Conv1D in Keras,
  # with the exception of kernel_range, bias_range, kernel_quantizer
  # and bias_quantizer, and kernel_initializer.
  #
  # kernel_quantizer: quantizer function/class for kernel
  # bias_quantizer: quantizer function/class for bias
  # kernel_range/bias_ranger: for quantizer functions whose values
  #   can go over [-1,+1], these values are used to set the clipping
  #   value of kernels and biases, respectively, instead of using the
  #   constraints specified by the user.
  #
  # we refer the reader to the documentation of Conv1D in Keras for the
  # other parameters.
  #

  def __init__(self,
               filters,
               kernel_size,
               strides=1,
               padding="valid",
               dilation_rate=1,
               activation=None,
               use_bias=True,
               kernel_initializer="he_normal",
               bias_initializer="zeros",
               kernel_regularizer=None,
               bias_regularizer=None,
               activity_regularizer=None,
               kernel_constraint=None,
               bias_constraint=None,
               kernel_quantizer=None,
               bias_quantizer=None,
               kernel_range=None,
               bias_range=None,
               **kwargs):

    if kernel_range is not None:
      warnings.warn("kernel_range is deprecated in QConv1D layer.")

    if bias_range is not None:
      warnings.warn("bias_range is deprecated in QConv1D layer.")

    self.kernel_range = kernel_range
    self.bias_range = bias_range

    self.kernel_quantizer = kernel_quantizer
    self.bias_quantizer = bias_quantizer

    self.kernel_quantizer_internal = get_quantizer(self.kernel_quantizer)
    self.bias_quantizer_internal = get_quantizer(self.bias_quantizer)

    # optimize parameter set to "auto" scaling mode if possible
    if hasattr(self.kernel_quantizer_internal, "_set_trainable_parameter"):
      self.kernel_quantizer_internal._set_trainable_parameter()

    self.quantizers = [
        self.kernel_quantizer_internal, self.bias_quantizer_internal
    ]

    kernel_constraint, kernel_initializer = (
        get_auto_range_constraint_initializer(self.kernel_quantizer_internal,
                                              kernel_constraint,
                                              kernel_initializer))

    if use_bias:
      bias_constraint, bias_initializer = (
          get_auto_range_constraint_initializer(self.bias_quantizer_internal,
                                                bias_constraint,
                                                bias_initializer))
    if activation is not None:
      activation = get_quantizer(activation)

    super(QConv1D, self).__init__(
        filters=filters,
        kernel_size=kernel_size,
        strides=strides,
        padding=padding,
        dilation_rate=dilation_rate,
        activation=activation,
        use_bias=use_bias,
        kernel_initializer=kernel_initializer,
        bias_initializer=bias_initializer,
        kernel_regularizer=kernel_regularizer,
        bias_regularizer=bias_regularizer,
        activity_regularizer=activity_regularizer,
        kernel_constraint=kernel_constraint,
        bias_constraint=bias_constraint,
        **kwargs)

  def call(self, inputs):
    if self.kernel_quantizer:
      quantized_kernel = self.kernel_quantizer_internal(self.kernel)
    else:
      quantized_kernel = self.kernel

    outputs = tf.keras.backend.conv1d(
        inputs,
        quantized_kernel,
        strides=self.strides[0],
        padding=self.padding,
        data_format=self.data_format,
        dilation_rate=self.dilation_rate[0])

    if self.use_bias:
      if self.bias_quantizer:
        quantized_bias = self.bias_quantizer_internal(self.bias)
      else:
        quantized_bias = self.bias

      outputs = tf.keras.backend.bias_add(
          outputs, quantized_bias, data_format=self.data_format)

    if self.activation is not None:
      return self.activation(outputs)
    return outputs

  def get_config(self):
    config = {
        "kernel_quantizer":
            constraints.serialize(self.kernel_quantizer_internal),
        "bias_quantizer":
            constraints.serialize(self.bias_quantizer_internal),
        "kernel_range": self.kernel_range,
        "bias_range": self.bias_range
    }
    base_config = super(QConv1D, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))

  def get_quantization_config(self):
    return {
        "kernel_quantizer":
            str(self.kernel_quantizer_internal),
        "bias_quantizer":
            str(self.bias_quantizer_internal),
        "activation":
            str(self.activation),
        "filters" : str(self.filters)
    }

  def get_quantizers(self):
    return self.quantizers

  def get_prunable_weights(self):
    return [self.kernel]


class QConv2D(Conv2D, PrunableLayer):
  """2D convolution layer (e.g. spatial convolution over images)."""

  # most of these parameters follow the implementation of Conv2D in Keras,
  # with the exception of kernel_range, bias_range, kernel_quantizer
  # and bias_quantizer, and kernel_initializer.
  #
  # kernel_quantizer: quantizer function/class for kernel
  # bias_quantizer: quantizer function/class for bias
  # kernel_range/bias_ranger: for quantizer functions whose values
  #   can go over [-1,+1], these values are used to set the clipping
  #   value of kernels and biases, respectively, instead of using the
  #   constraints specified by the user.
  #
  # we refer the reader to the documentation of Conv2D in Keras for the
  # other parameters.
  #

  def __init__(self,
               filters,
               kernel_size,
               strides=(1, 1),
               padding="valid",
               data_format="channels_last",
               dilation_rate=(1, 1),
               activation=None,
               use_bias=True,
               kernel_initializer="he_normal",
               bias_initializer="zeros",
               kernel_regularizer=None,
               bias_regularizer=None,
               activity_regularizer=None,
               kernel_constraint=None,
               bias_constraint=None,
               kernel_range=None,
               bias_range=None,
               kernel_quantizer=None,
               bias_quantizer=None,
               **kwargs):

    if kernel_range is not None:
      warnings.warn("kernel_range is deprecated in QConv2D layer.")

    if bias_range is not None:
      warnings.warn("bias_range is deprecated in QConv2D layer.")

    self.kernel_range = kernel_range
    self.bias_range = bias_range

    self.kernel_quantizer = kernel_quantizer
    self.bias_quantizer = bias_quantizer

    self.kernel_quantizer_internal = get_quantizer(self.kernel_quantizer)
    self.bias_quantizer_internal = get_quantizer(self.bias_quantizer)

    # optimize parameter set to "auto" scaling mode if possible
    if hasattr(self.kernel_quantizer_internal, "_set_trainable_parameter"):
      self.kernel_quantizer_internal._set_trainable_parameter()

    self.quantizers = [
        self.kernel_quantizer_internal, self.bias_quantizer_internal
    ]

    kernel_constraint, kernel_initializer = (
        get_auto_range_constraint_initializer(self.kernel_quantizer_internal,
                                              kernel_constraint,
                                              kernel_initializer))

    if use_bias:
      bias_constraint, bias_initializer = (
          get_auto_range_constraint_initializer(self.bias_quantizer_internal,
                                                bias_constraint,
                                                bias_initializer))

    if activation is not None:
      activation = get_quantizer(activation)

    super(QConv2D, self).__init__(
        filters=filters,
        kernel_size=kernel_size,
        strides=strides,
        padding=padding,
        data_format=data_format,
        dilation_rate=dilation_rate,
        activation=activation,
        use_bias=use_bias,
        kernel_initializer=kernel_initializer,
        bias_initializer=bias_initializer,
        kernel_regularizer=kernel_regularizer,
        bias_regularizer=bias_regularizer,
        activity_regularizer=activity_regularizer,
        kernel_constraint=kernel_constraint,
        bias_constraint=bias_constraint,
        **kwargs)

  def call(self, inputs):
    if self.kernel_quantizer:
      quantized_kernel = self.kernel_quantizer_internal(self.kernel)
    else:
      quantized_kernel = self.kernel

    outputs = tf.keras.backend.conv2d(
        inputs,
        quantized_kernel,
        strides=self.strides,
        padding=self.padding,
        data_format=self.data_format,
        dilation_rate=self.dilation_rate)

    if self.use_bias:
      if self.bias_quantizer:
        quantized_bias = self.bias_quantizer_internal(self.bias)
      else:
        quantized_bias = self.bias

      outputs = tf.keras.backend.bias_add(
          outputs, quantized_bias, data_format=self.data_format)

    if self.activation is not None:
      return self.activation(outputs)
    return outputs

  def get_config(self):
    config = {
        "kernel_quantizer":
            constraints.serialize(self.kernel_quantizer_internal),
        "bias_quantizer":
            constraints.serialize(self.bias_quantizer_internal),
        "kernel_range": self.kernel_range,
        "bias_range": self.bias_range
    }
    base_config = super(QConv2D, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))

  def get_quantization_config(self):
    return {
        "kernel_quantizer":
            str(self.kernel_quantizer_internal),
        "bias_quantizer":
            str(self.bias_quantizer_internal),
        "activation":
            str(self.activation),
        "filters" : str(self.filters)
    }

  def get_quantizers(self):
    return self.quantizers

  def get_prunable_weights(self):
    return [self.kernel]


class QConv2DTranspose(Conv2DTranspose, PrunableLayer):
  """2D convolution layer (e.g. spatial convolution over images)."""

  # most of these parameters follow the implementation of Conv2DTranspose
  # in Keras, with the exception of kernel_quantizer and bias_quantizer
  # and kernel_initializer.
  #
  # kernel_quantizer: quantizer function/class for kernel
  # bias_quantizer: quantizer function/class for bias
  #
  # we refer the reader to the documentation of Conv2DTranspose in Keras for
  # the other parameters.
  #

  def __init__(self,
               filters,
               kernel_size,
               strides=(1, 1),
               padding='valid',
               output_padding=None,
               data_format=None,
               dilation_rate=(1, 1),
               activation=None,
               use_bias=True,
               kernel_initializer='glorot_uniform',
               bias_initializer='zeros',
               kernel_regularizer=None,
               bias_regularizer=None,
               activity_regularizer=None,
               kernel_constraint=None,
               bias_constraint=None,
               kernel_quantizer=None,
               bias_quantizer=None,
               **kwargs):

    self.kernel_quantizer = kernel_quantizer
    self.bias_quantizer = bias_quantizer

    self.kernel_quantizer_internal = get_quantizer(self.kernel_quantizer)
    self.bias_quantizer_internal = get_quantizer(self.bias_quantizer)

    # optimize parameter set to "auto" scaling mode if possible
    if hasattr(self.kernel_quantizer_internal, "_set_trainable_parameter"):
      self.kernel_quantizer_internal._set_trainable_parameter()

    self.quantizers = [
        self.kernel_quantizer_internal, self.bias_quantizer_internal
    ]

    kernel_constraint, kernel_initializer = (
        get_auto_range_constraint_initializer(self.kernel_quantizer_internal,
                                              kernel_constraint,
                                              kernel_initializer))

    if use_bias:
      bias_constraint, bias_initializer = (
          get_auto_range_constraint_initializer(self.bias_quantizer_internal,
                                                bias_constraint,
                                                bias_initializer))

    if activation is not None:
      activation = get_quantizer(activation)

    super(QConv2DTranspose, self).__init__(
        filters=filters,
        kernel_size=kernel_size,
        strides=strides,
        padding=padding,
        output_padding=None,
        data_format=data_format,
        dilation_rate=dilation_rate,
        activation=activation,
        use_bias=use_bias,
        kernel_initializer=kernel_initializer,
        bias_initializer=bias_initializer,
        kernel_regularizer=kernel_regularizer,
        bias_regularizer=bias_regularizer,
        activity_regularizer=activity_regularizer,
        kernel_constraint=kernel_constraint,
        bias_constraint=bias_constraint,
        **kwargs)

  def call(self, inputs):
    inputs_shape = array_ops.shape(inputs)
    batch_size = inputs_shape[0]
    if self.data_format == 'channels_first':
      h_axis, w_axis = 2, 3
    else:
      h_axis, w_axis = 1, 2

    height, width = inputs_shape[h_axis], inputs_shape[w_axis]
    kernel_h, kernel_w = self.kernel_size
    stride_h, stride_w = self.strides

    if self.output_padding is None:
      out_pad_h = out_pad_w = None
    else:
      out_pad_h, out_pad_w = self.output_padding

    # Infer the dynamic output shape:
    out_height = conv_utils.deconv_output_length(height,
                                                 kernel_h,
                                                 padding=self.padding,
                                                 output_padding=out_pad_h,
                                                 stride=stride_h,
                                                 dilation=self.dilation_rate[0])
    out_width = conv_utils.deconv_output_length(width,
                                                kernel_w,
                                                padding=self.padding,
                                                output_padding=out_pad_w,
                                                stride=stride_w,
                                                dilation=self.dilation_rate[1])
    if self.data_format == 'channels_first':
      output_shape = (batch_size, self.filters, out_height, out_width)
    else:
      output_shape = (batch_size, out_height, out_width, self.filters)

    if self.kernel_quantizer:
      quantized_kernel = self.kernel_quantizer_internal(self.kernel)
    else:
      quantized_kernel = self.kernel

    output_shape_tensor = array_ops.stack(output_shape)
    outputs = tf.keras.backend.conv2d_transpose(
        inputs,
        quantized_kernel,
        output_shape_tensor,
        strides=self.strides,
        padding=self.padding,
        data_format=self.data_format,
        dilation_rate=self.dilation_rate)

    if not context.executing_eagerly():
      # Infer the static output shape:
      out_shape = self.compute_output_shape(inputs.shape)
      outputs.set_shape(out_shape)

    if self.use_bias:
      if self.bias_quantizer:
        quantized_bias = self.bias_quantizer_internal(self.bias)
      else:
        quantized_bias = self.bias

      outputs = tf.keras.backend.bias_add(
          outputs,
          quantized_bias,
          data_format=self.data_format)

    if self.activation is not None:
      return self.activation(outputs)
    return outputs

  def get_config(self):
    config = {
        "kernel_quantizer":
            constraints.serialize(self.kernel_quantizer_internal),
        "bias_quantizer":
            constraints.serialize(self.bias_quantizer_internal)
    }
    base_config = super(QConv2DTranspose, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))

  def get_quantizers(self):
    return self.quantizers

  def get_prunable_weights(self):
    return [self.kernel]


class QSeparableConv1D(SeparableConv1D, PrunableLayer):
  """Depthwise separable 1D convolution."""

  # most of these parameters follow the implementation of SeparableConv1D
  # in Keras, with the exception of depthwise_quantizer, pointwise_quantizer
  # and bias_quantizer.
  #
  # depthwise_quantizer: quantizer function/class for depthwise spatial kernel
  # pointwise_quantizer: quantizer function/class for pointwise kernel
  # bias_quantizer: quantizer function/class for bias
  #
  # we refer the reader to the documentation of SeparableConv1D in Keras for
  # the other parameters.
  #

  def __init__(self,
               filters,
               kernel_size,
               strides=1,
               padding='valid',
               data_format=None,
               dilation_rate=1,
               depth_multiplier=1,
               activation=None,
               use_bias=True,
               depthwise_initializer='glorot_uniform',
               pointwise_initializer='glorot_uniform',
               bias_initializer='zeros',
               depthwise_regularizer=None,
               pointwise_regularizer=None,
               bias_regularizer=None,
               activity_regularizer=None,
               depthwise_constraint=None,
               pointwise_constraint=None,
               bias_constraint=None,
               depthwise_quantizer=None,
               pointwise_quantizer=None,
               bias_quantizer=None,
               **kwargs):

    self.depthwise_quantizer = depthwise_quantizer
    self.pointwise_quantizer = pointwise_quantizer
    self.bias_quantizer = bias_quantizer

    self.depthwise_quantizer_internal = get_quantizer(self.depthwise_quantizer)
    self.pointwise_quantizer_internal = get_quantizer(self.pointwise_quantizer)
    self.bias_quantizer_internal = get_quantizer(self.bias_quantizer)

    # optimize parameter set to "auto" scaling mode if possible
    if hasattr(self.depthwise_quantizer_internal, "_set_trainable_parameter"):
      self.depthwise_quantizer_internal._set_trainable_parameter()

    if hasattr(self.pointwise_quantizer_internal, "_set_trainable_parameter"):
      self.pointwise_quantizer_internal._set_trainable_parameter()

    self.quantizers = [
        self.depthwise_quantizer_internal, self.pointwise_quantizer_internal,
        self.bias_quantizer_internal
    ]

    depthwise_constraint, depthwise_initializer = (
        get_auto_range_constraint_initializer(self.depthwise_quantizer_internal,
                                              depthwise_constraint,
                                              depthwise_initializer))

    pointwise_constraint, pointwise_initializer = (
        get_auto_range_constraint_initializer(self.pointwise_quantizer_internal,
                                              pointwise_constraint,
                                              pointwise_initializer))

    if use_bias:
      bias_constraint, bias_initializer = (
          get_auto_range_constraint_initializer(self.bias_quantizer_internal,
                                                bias_constraint,
                                                bias_initializer))

    if activation is not None:
      activation = get_quantizer(activation)

    super(QSeparableConv1D, self).__init__(
        filters=filters,
        kernel_size=kernel_size,
        strides=strides,
        padding=padding,
        data_format=data_format,
        dilation_rate=dilation_rate,
        depth_multiplier=depth_multiplier,
        activation=activation,
        use_bias=use_bias,
        depthwise_initializer=initializers.get(depthwise_initializer),
        pointwise_initializer=initializers.get(pointwise_initializer),
        bias_initializer=initializers.get(bias_initializer),
        depthwise_regularizer=regularizers.get(depthwise_regularizer),
        pointwise_regularizer=regularizers.get(pointwise_regularizer),
        bias_regularizer=regularizers.get(bias_regularizer),
        activity_regularizer=regularizers.get(activity_regularizer),
        depthwise_constraint=constraints.get(depthwise_constraint),
        pointwise_constraint=constraints.get(pointwise_constraint),
        bias_constraint=constraints.get(bias_constraint),
        **kwargs)

  def call(self, inputs):
    if self.padding == 'causal':
      inputs = array_ops.pad(inputs, self._compute_causal_padding())
    
    spatial_start_dim = 1 if self.data_format == 'channels_last' else 2

    # Explicitly broadcast inputs and kernels to 4D.
    inputs = array_ops.expand_dims(inputs, spatial_start_dim)
    depthwise_kernel = array_ops.expand_dims(self.depthwise_kernel, 0)
    pointwise_kernel = array_ops.expand_dims(self.pointwise_kernel, 0)
    dilation_rate = (1,) + self.dilation_rate

    if self.padding == 'causal':
      op_padding = 'valid'
    else:
      op_padding = self.padding

    if self.depthwise_quantizer:
      quantized_depthwise_kernel = self.depthwise_quantizer_internal(
          depthwise_kernel)
    else:
      quantized_depthwise_kernel = depthwise_kernel

    if self.pointwise_quantizer:
      quantized_pointwise_kernel = self.pointwise_quantizer_internal(
          pointwise_kernel)
    else:
      quantized_pointwise_kernel = pointwise_kernel

    outputs = tf.keras.backend.separable_conv2d(
        inputs,
        quantized_depthwise_kernel,
        quantized_pointwise_kernel,
        strides=self.strides * 2,
        padding=op_padding,
        dilation_rate=dilation_rate,
        data_format=self.data_format)

    if self.use_bias:
      if self.bias_quantizer:
        quantized_bias = self.bias_quantizer_internal(self.bias)
      else:
        quantized_bias = self.bias

      outputs = tf.keras.backend.bias_add(
          outputs,
          quantized_bias,
          data_format=self.data_format)

    outputs = array_ops.squeeze(outputs, [spatial_start_dim])

    if self.activation is not None:
      return self.activation(outputs)
    return outputs

  def get_config(self):
    config = {
        "depthwise_quantizer":
            constraints.serialize(self.depthwise_quantizer_internal),
        "pointwise_quantizer":
            constraints.serialize(self.pointwise_quantizer_internal),
        "bias_quantizer":
            constraints.serialize(self.bias_quantizer_internal)
    }
    base_config = super(QSeparableConv1D, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))

  def get_quantizers(self):
    return self.quantizers

  def get_prunable_weights(self):
    return [self.depthwise_kernel, self.pointwise_kernel]


class QSeparableConv2D(SeparableConv2D, PrunableLayer):
  """Depthwise separable 2D convolution."""

  # most of these parameters follow the implementation of SeparableConv2D
  # in Keras, with the exception of depthwise_quantizer, pointwise_quantizer
  # and bias_quantizer.
  #
  # depthwise_quantizer: quantizer function/class for depthwise spatial kernel
  # pointwise_quantizer: quantizer function/class for pointwise kernel
  # bias_quantizer: quantizer function/class for bias
  #
  # we refer the reader to the documentation of SeparableConv2D in Keras for
  # the other parameters.
  #

  def __init__(self,
               filters,
               kernel_size,
               strides=(1, 1),
               padding='valid',
               data_format=None,
               dilation_rate=(1, 1),
               depth_multiplier=1,
               activation=None,
               use_bias=True,
               depthwise_initializer='glorot_uniform',
               pointwise_initializer='glorot_uniform',
               bias_initializer='zeros',
               depthwise_regularizer=None,
               pointwise_regularizer=None,
               bias_regularizer=None,
               activity_regularizer=None,
               depthwise_constraint=None,
               pointwise_constraint=None,
               bias_constraint=None,
               depthwise_quantizer=None,
               pointwise_quantizer=None,
               bias_quantizer=None,
               **kwargs):

    self.depthwise_quantizer = depthwise_quantizer
    self.pointwise_quantizer = pointwise_quantizer
    self.bias_quantizer = bias_quantizer

    self.depthwise_quantizer_internal = get_quantizer(self.depthwise_quantizer)
    self.pointwise_quantizer_internal = get_quantizer(self.pointwise_quantizer)
    self.bias_quantizer_internal = get_quantizer(self.bias_quantizer)

    # optimize parameter set to "auto" scaling mode if possible
    if hasattr(self.depthwise_quantizer_internal, "_set_trainable_parameter"):
      self.depthwise_quantizer_internal._set_trainable_parameter()

    if hasattr(self.pointwise_quantizer_internal, "_set_trainable_parameter"):
      self.pointwise_quantizer_internal._set_trainable_parameter()

    self.quantizers = [
        self.depthwise_quantizer_internal, self.pointwise_quantizer_internal,
        self.bias_quantizer_internal
    ]

    depthwise_constraint, depthwise_initializer = (
        get_auto_range_constraint_initializer(self.depthwise_quantizer_internal,
                                              depthwise_constraint,
                                              depthwise_initializer))

    pointwise_constraint, pointwise_initializer = (
        get_auto_range_constraint_initializer(self.pointwise_quantizer_internal,
                                              pointwise_constraint,
                                              pointwise_initializer))

    if use_bias:
      bias_constraint, bias_initializer = (
          get_auto_range_constraint_initializer(self.bias_quantizer_internal,
                                                bias_constraint,
                                                bias_initializer))

    if activation is not None:
      activation = get_quantizer(activation)

    super(QSeparableConv2D, self).__init__(
        filters=filters,
        kernel_size=kernel_size,
        strides=strides,
        padding=padding,
        data_format=data_format,
        dilation_rate=dilation_rate,
        depth_multiplier=depth_multiplier,
        activation=activation,
        use_bias=use_bias,
        depthwise_initializer=initializers.get(depthwise_initializer),
        pointwise_initializer=initializers.get(pointwise_initializer),
        bias_initializer=initializers.get(bias_initializer),
        depthwise_regularizer=regularizers.get(depthwise_regularizer),
        pointwise_regularizer=regularizers.get(pointwise_regularizer),
        bias_regularizer=regularizers.get(bias_regularizer),
        activity_regularizer=regularizers.get(activity_regularizer),
        depthwise_constraint=constraints.get(depthwise_constraint),
        pointwise_constraint=constraints.get(pointwise_constraint),
        bias_constraint=constraints.get(bias_constraint),
        **kwargs)

  def call(self, inputs):
    # Apply the actual ops.
    if self.depthwise_quantizer:
      quantized_depthwise_kernel = self.depthwise_quantizer_internal(
          self.depthwise_kernel)
    else:
      quantized_depthwise_kernel = self.depthwise_kernel

    if self.pointwise_quantizer:
      quantized_pointwise_kernel = self.pointwise_quantizer_internal(
          self.pointwise_kernel)
    else:
      quantized_pointwise_kernel = self.pointwise_kernel

    outputs = tf.keras.backend.separable_conv2d(
        inputs,
        quantized_depthwise_kernel,
        quantized_pointwise_kernel,
        strides=self.strides,
        padding=self.padding,
        dilation_rate=self.dilation_rate,
        data_format=self.data_format)

    if self.use_bias:
      if self.bias_quantizer:
        quantized_bias = self.bias_quantizer_internal(self.bias)
      else:
        quantized_bias = self.bias

      outputs = tf.keras.backend.bias_add(
          outputs,
          quantized_bias,
          data_format=self.data_format)

    if self.activation is not None:
      return self.activation(outputs)
    return outputs

  def get_config(self):
    config = {
        "depthwise_quantizer":
            constraints.serialize(self.depthwise_quantizer_internal),
        "pointwise_quantizer":
            constraints.serialize(self.pointwise_quantizer_internal),
        "bias_quantizer":
            constraints.serialize(self.bias_quantizer_internal)
    }
    base_config = super(QSeparableConv2D, self).get_config()
    return dict(list(base_config.items()) + list(config.items()))

  def get_quantizers(self):
    return self.quantizers

  def get_prunable_weights(self):
    return [self.depthwise_kernel, self.pointwise_kernel]


class QDepthwiseConv2D(DepthwiseConv2D, PrunableLayer):
  """Creates quantized depthwise conv2d. Copied from mobilenet."""

  # most of these parameters follow the implementation of DepthwiseConv2D
  # in Keras, # with the exception of depthwise_range, bias_range,
  # depthwise_quantizer # and bias_quantizer, and kernel_initializer.
  #
  # depthwise_quantizer: quantizer function/class for kernel
  # bias_quantizer: quantizer function/class for bias
  # depthwise_range/bias_ranger: for quantizer functions whose values
  #   can go over [-1,+1], these values are used to set the clipping
  #   value of kernels and biases, respectively, instead of using the
  #   constraints specified by the user.
  #
  # we refer the reader to the documentation of DepthwiseConv2D in Keras for the
  # other parameters.
  #

  def __init__(self,
               kernel_size,
               strides=(1, 1),
               padding="VALID",
               depth_multiplier=1,
               data_format=None,
               activation=None,
               use_bias=True,
               depthwise_initializer="he_normal",
               bias_initializer="zeros",
               depthwise_regularizer=None,
               bias_regularizer=None,
               activity_regularizer=None,
               depthwise_constraint=None,
               bias_constraint=None,
               dilation_rate=(1, 1),
               depthwise_quantizer=None,
               bias_quantizer=None,
               depthwise_range=None,
               bias_range=None,
               **kwargs):

    if depthwise_range is not None:
      warnings.warn("depthwise_range is deprecated in QDepthwiseConv2D layer.")

    if bias_range is not None:
      warnings.warn("bias_range is deprecated in QDepthwiseConv2D layer.")

    self.depthwise_range = depthwise_range
    self.bias_range = bias_range

    self.depthwise_quantizer = depthwise_quantizer
    self.bias_quantizer = bias_quantizer

    self.depthwise_quantizer_internal = get_quantizer(self.depthwise_quantizer)
    self.bias_quantizer_internal = get_quantizer(self.bias_quantizer)

    # optimize parameter set to "auto" scaling mode if possible
    if hasattr(self.depthwise_quantizer_internal, "_set_trainable_parameter"):
      self.depthwise_quantizer_internal._set_trainable_parameter()

    self.quantizers = [
        self.depthwise_quantizer_internal, self.bias_quantizer_internal
    ]

    depthwise_constraint, depthwise_initializer = (
        get_auto_range_constraint_initializer(self.depthwise_quantizer_internal,
                                              depthwise_constraint,
                                              depthwise_initializer))

    if use_bias:
      bias_constraint, bias_initializer = (
          get_auto_range_constraint_initializer(self.bias_quantizer_internal,
                                                bias_constraint,
                                                bias_initializer))
    if activation is not None:
      activation = get_quantizer(activation)

    super(QDepthwiseConv2D, self).__init__(
        kernel_size=kernel_size,
        strides=strides,
        padding=padding,
        data_format=data_format,
        activation=activation,
        use_bias=use_bias,
        depthwise_regularizer=depthwise_regularizer,
        bias_regularizer=bias_regularizer,
        activity_regularizer=activity_regularizer,
        depth_multiplier=depth_multiplier,
        depthwise_initializer=depthwise_initializer,
        bias_initializer=bias_initializer,
        depthwise_constraint=depthwise_constraint,
        bias_constraint=bias_constraint,
        dilation_rate=dilation_rate,
        **kwargs)

  def build(self, input_shape):
    if len(input_shape) < 4:
      raise ValueError(
          "Inputs to `QDepthwiseConv2D` should have rank 4. "
          "Received input shape:", str(input_shape))
    if self.data_format == "channels_first":
      channel_axis = 1
    else:
      channel_axis = 3
    if input_shape[channel_axis] is None:
      raise ValueError("The channel dimension of the inputs to "
                       "`QDepthwiseConv2D` "
                       "should be defined. Found `None`.")
    input_dim = int(input_shape[channel_axis])
    depthwise_kernel_shape = (self.kernel_size[0], self.kernel_size[1],
                              input_dim, self.depth_multiplier)

    self.depthwise_kernel = self.add_weight(
        shape=depthwise_kernel_shape,
        initializer=self.depthwise_initializer,
        name="depthwise_kernel",
        regularizer=self.depthwise_regularizer,
        constraint=self.depthwise_constraint)

    if self.use_bias:
      self.bias = self.add_weight(
          shape=(input_dim * self.depth_multiplier,),
          initializer=self.bias_initializer,
          name="bias",
          regularizer=self.bias_regularizer,
          constraint=self.bias_constraint)
    else:
      self.bias = None
    # Set input spec.
    self.input_spec = InputSpec(ndim=4, axes={channel_axis: input_dim})
    self.built = True

  def call(self, inputs, training=None):
    if self.depthwise_quantizer:
      quantized_depthwise_kernel = (
          self.depthwise_quantizer_internal(self.depthwise_kernel))
    else:
      quantized_depthwise_kernel = self.depthwise_kernel
    outputs = tf.keras.backend.depthwise_conv2d(
        inputs,
        quantized_depthwise_kernel,
        strides=self.strides,
        padding=self.padding,
        dilation_rate=self.dilation_rate,
        data_format=self.data_format)

    if self.use_bias:
      if self.bias_quantizer:
        quantized_bias = self.bias_quantizer_internal(self.bias)
      else:
        quantized_bias = self.bias
      outputs = tf.keras.backend.bias_add(
          outputs, quantized_bias, data_format=self.data_format)

    if self.activation is not None:
      return self.activation(outputs)

    return outputs

  def get_config(self):
    config = super(QDepthwiseConv2D, self).get_config()
    config.pop("filters", None)
    config.pop("kernel_initializer", None)
    config.pop("kernel_regularizer", None)
    config.pop("kernel_constraint", None)
    config["depth_multiplier"] = self.depth_multiplier
    config["depthwise_initializer"] = initializers.serialize(
        self.depthwise_initializer)
    config["depthwise_regularizer"] = regularizers.serialize(
        self.depthwise_regularizer)
    config["depthwise_constraint"] = constraints.serialize(
        self.depthwise_constraint)
    config["depthwise_quantizer"] = constraints.serialize(
        self.depthwise_quantizer_internal)
    config["bias_quantizer"] = constraints.serialize(
        self.bias_quantizer_internal)
    config["depthwise_range"] = self.depthwise_range
    config["bias_range"] = self.bias_range
    return config

  def get_quantization_config(self):
    return {
        "depthwise_quantizer_internal":
            str(self.depthwise_quantizer_internal),
        "bias_quantizer":
            str(self.bias_quantizer_internal),
        "activation":
            str(self.activation),
        "filters" : str(self.filters)
    }

  def get_quantizers(self):
    return self.quantizers

  def get_prunable_weights(self):
    return [self.depthwise_kernel]


def QMobileNetSeparableConv2D(
    filters,  # pylint: disable=invalid-name
    kernel_size,
    strides=(1, 1),
    padding="VALID",
    dilation_rate=(1, 1),
    depth_multiplier=1,
    activation=None,
    use_bias=True,
    depthwise_initializer="he_normal",
    pointwise_initializer="he_normal",
    bias_initializer="zeros",
    depthwise_regularizer=None,
    pointwise_regularizer=None,
    bias_regularizer=None,
    activity_regularizer=None,
    depthwise_constraint=None,
    pointwise_constraint=None,
    bias_constraint=None,
    depthwise_quantizer=None,
    pointwise_quantizer=None,
    bias_quantizer=None,
    depthwise_activation=None,
    depthwise_range=None,
    pointwise_range=None,
    bias_range=None,
    depthwise_dropout_rate=0.0,
    pw_first=False,
    name=""):
  """Adds a quantized separableconv2d."""

  # we use here a modified version that appeared in mobilenet that adds
  # quantization to the network, and possibly an intermediate activation
  # layer that acts as a quantizer and possible dropout layer between
  # the depthwise and pointwise convolutions.
  #
  # since this implementation expands into depthwise -> pointwise
  # convolutions, the users will not see a separable convolution operation
  # in model.summary(), but rather a depthwise convolution followed by a
  # pointwise convolution.
  #
  # depthwise_quantizer: depthwise quantization function
  # pointwise_quantizer: pointwise quantization function
  # bias_quantizer: bias quantization function for the pointwise convolution
  # depthwise_range/pointwise_range/bias_range: ranges to be used if
  # quantization values can become greater than -1 and +1.
  # depthwise_dropout_rate: dropout between depthwise and pointwise is added
  #   if rate > 0.0
  # pw_first: this may disappear in the future, but as deep quantized networks
  #   sometimes behave in different ways, if we are using binary or ternary
  #   quantization, it may be better to apply pointwise before depthwise.
  #
  # For the remaining parameters, please refer to Keras implementation of
  # SeparableConv2D.
  #

  def _call(inputs):  # pylint: disable=invalid-name
    """Internally builds qseparableconv2d."""

    x = inputs

    if pw_first:
      x = QConv2D(
          filters, (1, 1),
          strides=(1, 1),
          padding="same",
          use_bias=use_bias,
          kernel_constraint=pointwise_constraint,
          kernel_initializer=pointwise_initializer,
          kernel_regularizer=pointwise_regularizer,
          kernel_quantizer=pointwise_quantizer,
          bias_quantizer=bias_quantizer,
          bias_regularizer=bias_regularizer,
          bias_initializer=bias_initializer,
          bias_constraint=bias_constraint,
          activity_regularizer=activity_regularizer,
          kernel_range=pointwise_range,
          bias_range=bias_range,
          name=name + "_pw")(
              x)

      if depthwise_activation:
        if isinstance(depthwise_activation, QActivation):
          x = depthwise_activation(x)
        else:
          x = QActivation(depthwise_activation, name=name + "_dw_act")(x)

      if depthwise_dropout_rate > 0.0:
        x = Dropout(rate=depthwise_dropout_rate, name=name + "_dw_dropout")(x)

    x = QDepthwiseConv2D(
        kernel_size,
        strides=strides,
        dilation_rate=dilation_rate,
        padding=padding,
        depth_multiplier=depth_multiplier,
        use_bias=False,
        depthwise_regularizer=depthwise_regularizer,
        depthwise_initializer=depthwise_initializer,
        depthwise_constraint=depthwise_constraint,
        depthwise_quantizer=depthwise_quantizer,
        depthwise_range=depthwise_range,
        name=name + "_dw")(
            x)

    if not pw_first:
      if depthwise_activation:
        if isinstance(depthwise_activation, QActivation):
          x = depthwise_activation(x)
        else:
          x = QActivation(depthwise_activation, name=name + "_dw_act")(x)

      if depthwise_dropout_rate > 0.0:
        x = Dropout(rate=depthwise_dropout_rate, name=name + "_dw_dropout")(x)

      x = QConv2D(
          filters, (1, 1),
          strides=(1, 1),
          padding="same",
          use_bias=use_bias,
          kernel_constraint=pointwise_constraint,
          kernel_initializer=pointwise_initializer,
          kernel_regularizer=pointwise_regularizer,
          kernel_quantizer=pointwise_quantizer,
          bias_quantizer=bias_quantizer,
          bias_regularizer=bias_regularizer,
          bias_initializer=bias_initializer,
          bias_constraint=bias_constraint,
          activity_regularizer=activity_regularizer,
          kernel_range=pointwise_range,
          bias_range=bias_range,
          name=name + "_pw")(
              x)

    if activation:
      if isinstance(activation, QActivation):
        x = activation(x)
      else:
        x = Activation(activation, name=name + "_pw_act")(x)
    return x

  return _call



# Lint as: python3
# ==============================================================================
# Copyright 2020 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Implements forgiving factor metrics."""

import numpy as np


class ForgivingFactor:
  """Base class. Should never be invoked."""

  def __init__(self, delta_p, delta_n, rate):
    self.delta_p = np.float32(delta_p) / 100.0
    self.delta_n = np.float32(delta_n) / 100.0
    self.rate = np.float32(rate)

  def get_reference(self, model):
    """Computes reference size of model."""

    raise Exception("class not implemented.")

  def get_trial(self, model, schema):
    """Computes size of quantization trial."""

    raise Exception("class not implemented.")

  def delta(self):
    return np.where(
        self.trial_size < self.reference_size,
        self.delta_p * (np.log(self.reference_size/self.trial_size) /
                        np.log(self.rate)),
        self.delta_n * (np.log(self.reference_size/self.trial_size) /
                        np.log(self.rate)))




# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""create merge layer output quantizers."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc

from qkeras.qtools.quantized_operators import adder_impl
from qkeras.qtools.quantized_operators import multiplier_factory
from qkeras.qtools.quantized_operators import quantizer_impl


class MergeFactory:
  """determine which merge implementation to use."""

  def make_quantizer(self, input_qe_list, layer_type):
    """make quantier."""

    if layer_type == "Add":
      return Add(input_qe_list)
    elif layer_type == "Multiply":
      return Multiply(input_qe_list)
    elif layer_type == "Maximum":
      return Maximum(input_qe_list)
    elif layer_type == "Minimum":
      return Minimum(input_qe_list)
    elif layer_type == "Average":
      return Average(input_qe_list)
    elif layer_type == "Concatenate":
      return Concatenate(input_qe_list)
    elif layer_type == "Dot":
      return Dot(input_qe_list)


class IMerger(abc.ABC):
  """abstract class for merge quantizer."""

  def __init__(self, input_qe_list):
    self.input_quantizers = []
    self.edges = []

    for node in input_qe_list:
      self.input_quantizers.append(node[0])
      self.edges.append(node[1])


class Add(IMerger):
  """add a list of inputs."""

  # It takes as input a list of tensors, all of the same shape,
  # and returns a single tensor (also of the same shape).

  def __init__(self, input_qe_list):
    super().__init__(input_qe_list)

    max_bits = -1
    max_int_bits = -1
    is_signed = False

    bits = 0
    is_floating_point = False
    for quantizer in self.input_quantizers:
      if quantizer.is_floating_point:
        is_floating_point = True
        bits = max(bits, quantizer.bits)
      else:
        if quantizer.is_po2:
          qbits_quantizer = adder_impl.po2_qbits_converter(
              quantizer)
        else:
          qbits_quantizer = quantizer

        if qbits_quantizer.bits > max_bits:
          max_bits = qbits_quantizer.bits

        if qbits_quantizer.int_bits > max_int_bits:
          max_int_bits = qbits_quantizer.int_bits

      is_signed |= quantizer.is_signed

    if is_floating_point:
      self.output = quantizer_impl.FloatingPoint(
          bits=bits)
    else:
      self.output = quantizer_impl.QuantizedBits()
      self.output.bits = max_bits + 1
      self.output.int_bits = max_int_bits + 1
      self.output.is_signed = is_signed
      self.output.mode = 0
      self.output.is_floating_point = False
      self.output.is_po2 = 0

    self.gate_factor = 1
    self.gate_bits = self.output.bits

  def implemented_as(self):
    return "add"


class Multiply(IMerger):
  """multiplies (element-wise) a list of inputs."""

  # It takes as input a list of tensors, all of the same shape,
  # and returns a single tensor (also of the same shape).

  def  __init__(self, input_qe_list):
    super().__init__(input_qe_list)
    multiplier_instance = multiplier_factory.MultiplierFactory()

    quantizer = self.input_quantizers[0]
    for cur in self.input_quantizers[1:]:
      tmp = multiplier_instance.make_multiplier(quantizer, cur)
      quantizer = tmp.output

    self.output = quantizer

    # TODO(lishanok): only use the last multiplier here
    self.impl_class = tmp
    self.gate_factor = tmp.gate_factor
    self.gate_bits = tmp.gate_bits

  def implemented_as(self):
    return self.impl_class.implemented_as()


class Maximum(IMerger):
  """maximum of a list of inputs."""

  # It takes as input a list of tensors, all of the same shape,
  # and returns a single tensor (also of the same shape).

  def __init__(self, input_qe_list):
    super().__init__(input_qe_list)

    is_same = True
    is_floating_point = False
    bits = 0

    quantizer = self.input_quantizers[0]
    for cur in self.input_quantizers[1:]:
      if (quantizer.name != cur.name or quantizer.bits != cur.bits or
          quantizer.int_bits != cur.int_bits or
          quantizer.is_signed != cur.is_signed):
        is_same = False
        break

    if is_same:
      self.output = quantizer
    else:
      max_bits = -1
      max_int_bits = -1
      is_signed = False
      for quantizer in self.input_quantizers:
        if quantizer.is_floating_point:
          is_floating_point = True
          bits = max(bits, quantizer.bits)
        else:
          if quantizer.is_po2:
            qbits_quantizer = adder_impl.po2_qbits_converter(
                quantizer)
          else:
            qbits_quantizer = quantizer

          if qbits_quantizer.bits > max_bits:
            max_bits = qbits_quantizer.bits

          if qbits_quantizer.int_bits > max_int_bits:
            max_int_bits = qbits_quantizer.int_bits

        is_signed |= quantizer.is_signed

      if is_floating_point:
        self.output = quantizer_impl.FloatingPoint(
            bits=bits)
      else:
        self.output = quantizer_impl.QuantizedBits()
        self.output.bits = max_bits
        self.output.int_bits = max_int_bits
        self.output.is_signed = is_signed
        self.output.mode = 0
        self.output.is_floating_point = False
        self.output.is_po2 = 0

    self.gate_factor = 0.2
    self.gate_bits = self.output.bits

  @staticmethod
  def implemented_as():
    return "add"


class Minimum(Maximum):
  """minimum (element-wise) a list of inputs."""

  # It takes as input a list of tensors, all of the same shape,
  # and returns a single tensor (also of the same shape).
  pass


class Average(Maximum):
  """average (element-wise) a list of inputs."""

  # It takes as input a list of tensors, all of the same shape,
  # and returns a single tensor (also of the same shape).
  def __init__(self, input_qe_list):
    super().__init__(input_qe_list)

    self.gate_factor = 1
    self.gate_bits = self.output.bits


class Concatenate(Maximum):
  """Layer that concatenates a list of inputs."""

  # It takes as input a list of tensors, all of the same
  # shape except for the concatenation axis, and returns
  # a single tensor, the concatenation of all inputs..
  def __init__(self, input_qe_list):
    super().__init__(input_qe_list)

    self.gate_factor = 0
    self.gate_bits = self.output.bits


# TODO(lishanok): finish DOT ndimension tensor logic
class Dot(IMerger):
  """dot product between samples in two tensors."""

  # E.g. if applied to a list of two tensors a and b
  # of shape (batch_size, n), the
  # output will be a tensor of shape (batch_size, 1)
  # where each entry i will be\
  # the dot product between a[i] and b[i].

  pass



# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""configurations."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np


class ConfigClass:
  """configuration class."""

  def __init__(self):

    self.default_source_quantizer = "quantized_bits(8, 0, 1)"
    self.default_interm_quantizer = "fp32"

    # Horowitz estimates from ISSCC 2014

    self.fpm_add = np.poly1d([0.003125, 0])
    self.fpm_mul = np.poly1d([0.002994791667, 0.001041666667, 0])
    self.fp16_add = np.poly1d([0.4])
    self.fp16_mul = np.poly1d([1.1])
    self.fp32_add = np.poly1d([0.9])
    self.fp32_mul = np.poly1d([3.7])

    self.sram_rd = np.poly1d([0.02455, -0.2656, 0.8661])
    self.dram_rd = np.poly1d([20.3125, 0])
    self.sram_mul_factor = 1/64.
    self.dram_mul_factor = 1.0

    self.include_energy = {}
    self.include_energy["default"] = ["inputs", "parameters", "op_cost"]
    self.include_energy["QActivation"] = ["outputs"]
    self.include_energy["QAdaptiveActivation"] = ["outputs"]
    self.include_energy["Activation"] = ["outputs"]
    self.include_energy["QBatchNormalization"] = ["parameters"]
    self.include_energy["BatchNormalization"] = ["parameters"]
    self.include_energy["Add"] = ["op_cost"]
    self.include_energy["Subtract"] = ["op_cost"]
    self.include_energy["MaxPooling2D"] = ["op_cost"]
    self.include_energy["default"] = ["inputs", "parameters", "op_cost"]

  def update(self, process, cfg_setting):
    """update config."""

    # pylint: disable=bare-except
    try:
      self.default_source_quantizer = cfg_setting[
          "default_source_quantizer"]
    except:
      pass

    try:
      self.default_interm_quantizer = cfg_setting[
          "default_interm_quantizer"]
    except:
      pass

    try:
      self.fpm_add = np.poly1d(cfg_setting[process]["fpm_add"])
    except:
      pass

    try:
      self.fpm_mul = np.poly1d(cfg_setting[process]["fpm_mul"])
    except:
      pass

    try:
      self.fp16_add = np.poly1d(cfg_setting[process]["fp16_add"])
    except:
      pass

    try:
      self.fp16_mul = np.poly1d(cfg_setting[process]["fp16_mul"])
    except:
      pass

    try:
      self.fp32_add = np.poly1d(cfg_setting[process]["fp32_add"])
    except:
      pass

    try:
      self.fp32_mul = np.poly1d(cfg_setting[process]["fp32_mul"])
    except:
      pass

    try:
      self.sram_rd = np.poly1d(cfg_setting[process]["sram_rd"])
    except:
      pass

    try:
      self.dram_rd = np.poly1d(cfg_setting[process]["dram_rd"])
    except:  # pylint: disable=broad-except
      pass

    try:
      for key in cfg_setting["include_energy"]:
        self.include_energy[key] = cfg_setting["include_energy"][key]
        if "Q" == key[0]:
	        # use the same rule for keras layer and qkeras layer
          self.include_energy[key[1:]] = cfg_setting["include_energy"][key]
    except:
      pass


cfg = ConfigClass()




# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Generates MAC, input and output datatype for a qkeras model."""
import collections
import copy
import numpy as np

import networkx as nx
from qkeras.qtools import qgraph
from qkeras.qtools import qtools_util
from qkeras.qtools import quantized_operators
from qkeras.qtools.quantized_operators import quantizer_factory as quantizer_factory_module
from qkeras.qtools.settings import cfg


class TagMissingError(ValueError):
  pass


LayerDataType = collections.namedtuple(
    "LayerDataType",
    [
        "input_quantizer_list",
        "multiplier",
        "accumulator",

        "weight_quantizer",
        "w_shapes",

        "bias_quantizer",
        "b_shapes",

        "output_quantizer",
        "output_shapes",

        "operation_count",
    ],
)

QKERAS_LAYERS = [
    "QDense",
    "QConv1D",
    "QConv2D",
    "QDepthwiseConv2D",
    "QSeparableConv2D",
    "QOctaveConv2D",
]

KERAS_LAYERS = [
    "Dense",
    "Conv1D",
    "Conv2D",
    "DepthwiseConv2D",
    "SeparableConv2D",
    "OctaveConv2D",
]


def update_output_quantizer_in_graph(graph, node_id, quantizer_factory,
                                     new_quantizer, for_reference):
  """update the edge with output quantizer type."""

  node = graph.nodes[node_id]
  qkeras_output_quantizer = node["out_quantizer"]

  # If existing graph doesn't have a valid output quantizer
  # update graph with the new quantizer
  if (for_reference or not qkeras_output_quantizer or
      not quantizer_factory.is_quantizer_supported(qkeras_output_quantizer)):
    qkeras_output_quantizer = new_quantizer
    qgraph.GraphUpdateEdge(graph, node_id, qkeras_output_quantizer)

  # If activation specified, convert activation quantizer to qtools quantizer
  # If activation not secified, convert the new quantizer to qtools quantizer
  output_quantizer = quantizer_factory.make_quantizer(qkeras_output_quantizer)

  # Output_quantizer is used for updating dictionary in json
  return output_quantizer


def generate_layer_data_type_map(graph, source_quantizer_list, is_inference,
                                 keras_quantizer=None, keras_accumulator=None,
                                 for_reference=False, debug=False):
  """main funciton to generate datatype for each layer.

  For each type of layer, this function calculates the sizes and minimum
  number of bits required to represent the parameters and variables (e.g.,
  weights, bias, multiplier and accumulator - MAC, etc.) embedded in
  these layers.

  Args:
    graph: input graph that traverses the model
    source_quantizer_list: a list of quantizers for model inputs
    is_inference: whether model is pre-trained with weights available
    keras_quantizer: default quantizer used to quantize weights and bias
    keras_accumulator: default MAC quantizer to quantize multiplier,
      accumulator and output
    for_reference: whether to generate a map for a baseline model
    debug: whether to print debug messages

  Returns:
    a result containing the following fields:
    source_quantizer_list similar as input
    output_layers: names of the layers that are output layers
    input_layers: names of the layers that are input_layers,
    layer_data_type_map: data type map of each layer
  """

  quantizer_factory = quantizer_factory_module.QuantizerFactory()
  layer_data_type_map = collections.OrderedDict()

  # get the output layers

  output_layers = []
  input_layers = []
  predecessors = list(graph.predecessors(qgraph.SINK))
  successors = list(graph.successors(qgraph.SOURCE))

  for u in predecessors:
    if u == qgraph.SOURCE or u == qgraph.SINK:
      continue
    output_layers.append(graph.nodes[u]["layer"][0])

  for u in successors:
    if u == qgraph.SOURCE or u == qgraph.SINK:
      continue
    input_layers.append(graph.nodes[u]["layer"][0])

  for node_id in nx.topological_sort(graph):
    node = graph.nodes[node_id]
    node_type = node["type"][-1]
    layer = node["layer"][0]
    is_input_layer = layer in input_layers

    w_shapes = None
    b_shapes = None
    output_shapes = None

    if hasattr(layer, "output_shape"):
      output_shapes = layer.output_shape

    if hasattr(layer, "get_weights"):
      weights = layer.get_weights()
      if len(weights) != 0:
        w_shapes = layer.get_weights()[0].shape
        b_shapes = weights[0].shape[-1]

    if debug:
      print("########")
      if layer is not None:
        print(layer.name)
      else:
        print("None")

    # Deals with keras layer or lack of input quantizer in qkeras layer.
    input_qe_list = qtools_util.get_input_quantizers_advanced(
        graph, node_id, is_input_layer, quantizer_factory, cfg)

    if input_qe_list and node_id != qgraph.SINK:
      input_quantizer_list = []
      for node in input_qe_list:
        input_quantizer_list.append(node[0])

      # Calculates number of operations (multiplication/accumulation).
      # Previously Merge layers's inputs all have the same shape, however, in
      # MobilenetV3 we found that there is shape broadcast in the keras
      # Multiply layer. Therefore we use the shape with max size as the
      # input shape
      if len(input_qe_list) > 0:
        maxsize = -1
        max_id = 0
        for (idx, item) in enumerate(input_qe_list):
          shape = item[1]["shape"]
          size = np.prod(shape[1:])
          if size > maxsize:
            maxsize = size
            max_id = idx
        input_shape = input_qe_list[max_id][1]["shape"]
      else:
        (_, edge_0) = input_qe_list[0]
        input_shape = edge_0["shape"]

      operation_count = qtools_util.get_operation_count(
          layer, input_shape)

    # Merges layers with multiple inputs.
    if qtools_util.is_merge_layers(layer):

      # merge_factory.make_quantizer automatically calculates the merge output
      # quantizer bitwidth according to input quantizer type.
      merge_factory = quantized_operators.MergeFactory()
      merge_quantizer = merge_factory.make_quantizer(
          input_qe_list, layer.__class__.__name__)

      if for_reference:
        # The for_reference option overwrites the auto-calculated merge output
        # quantizer
        if keras_accumulator:
          # gate_factor and gate_bits remain the same as previously
          # calculated; only change output quantizer as the keras_accumulator
          merge_quantizer.output = quantizer_factory.make_default_quantizer(
              mode=keras_accumulator)
        else:
          merge_quantizer.output = quantizer_factory.make_default_quantizer(
              mode=cfg.default_interm_quantizer)

      output_quantizer = update_output_quantizer_in_graph(
          graph, node_id, quantizer_factory, merge_quantizer.output,
          for_reference)

      layer_data_type_map[layer] = LayerDataType(
          input_quantizer_list,
          merge_quantizer,
          None,

          None,
          None,

          None,
          None,

          output_quantizer,
          output_shapes,

          operation_count
      )

    # MaxPooling/reshape/flatten/UpSampling1D/2D/3D
    elif (qtools_util.is_shape_alternation_layers(layer) or
          "UpSampling" in layer.__class__.__name__):
      input_quantizer = input_quantizer_list[0]

      # Output quantizer
      output_quantizer = update_output_quantizer_in_graph(
          graph, node_id, quantizer_factory, input_quantizer, for_reference)

      layer_data_type_map[layer] = LayerDataType(
          input_quantizer_list,
          None,
          None,

          None,
          None,

          None,
          None,

          output_quantizer,
          output_shapes,

          operation_count
      )

    # AveragePooling and GlobalAveragePooling
    elif layer.__class__.__name__ in [
        "AveragePooling2D", "AvgPool2D", "GlobalAvgPool2D",
        "GlobalAveragePooling2D", "QAveragePooling2D",
        "QGlobalAveragePooling2D"]:
      (input_quantizer, _) = input_qe_list[0]

      # This is a hack. We don't want to implement a new accumulator class
      # just for averagpooling. So we re-use accumulator type in conv/dense
      # layers which need multiplier and kernel as input parameters.
      # In order to do so, we fake a multiplier which treat the pool_size as
      # the kernel. since kernel needs 4 dimension, k_h, k_w, C_in, C_out,
      # we set the last two dimension as [1, 1]
      if layer.__class__.__name__ in ["AveragePooling2D", "AvgPool2D",
                                      "QAveragePooling2D"]:
        pool_size = tuple(list(layer.pool_size) + [1, 1])
      else:
        pool_size = tuple(list(input_shape)[1:-1] + [1, 1])

      # Automatically calculates the accumulator bitwidth according to input
      # quantizer type for both quantized pooling and regular pooling layers
      multiplier_factory = quantized_operators.MultiplierFactory()
      fake_multiplier = multiplier_factory.make_multiplier(
          input_quantizer, input_quantizer)
      fake_multiplier.output = input_quantizer
      accumulator_factory = quantized_operators.AccumulatorFactory()
      accumulator = accumulator_factory.make_accumulator(
          pool_size, fake_multiplier, use_bias=False)

      # For quantized pooling layers, we also need to consider the division
      # precision, which is controlled by the average quantizer
      if layer.__class__.__name__ in ["QAveragePooling2D",
                                      "QGlobalAveragePooling2D"]:
        # For the quantized layer, there is an average_quantizer used for
        # the inverse of division operation.
        qkeras_average_quantizer = layer.get_quantizers()[0]
        qtools_average_quantizer = quantizer_factory.make_quantizer(
            qkeras_average_quantizer)
        multiplier = multiplier_factory.make_multiplier(
            accumulator.output, qtools_average_quantizer)
      else:
        multiplier = None
      if debug:
        print("accumulator:", accumulator.output.bits)

      # Re-calcualte accumulator/multiplier type when it's using
      # for_reference option
      if for_reference:
        if keras_accumulator:
          # If keras_accumulator exists, use keras_accumulator as multiplier
          # or accumulator type
          if multiplier:
            # Quantized layers need to define multiplier type
            multiplier.output = quantizer_factory.make_default_quantizer(
                mode=keras_accumulator)
          accumulator.output = quantizer_factory.make_default_quantizer(
              mode=keras_accumulator)
        else:
          # If user didn't provide keras_accumulator, use the default settings
          # in cfg to define multiplier/accumulator type
          if multiplier:
            multiplier.output = quantizer_factory.make_default_quantizer(
                mode=cfg.default_interm_quantizer)
          accumulator.output = quantizer_factory.make_default_quantizer(
              mode=cfg.default_interm_quantizer)
        layer_quantizer = accumulator.output

      # set the output quantizer
      if layer.__class__.__name__ in ["QAveragePooling2D",
                                      "QGlobalAveragePooling2D"]:
        # If is quantized layer, last operation is multiply (averaging).
        layer_quantizer = multiplier.output
      else:
        layer_quantizer = accumulator.output
      output_quantizer = update_output_quantizer_in_graph(
          graph, node_id, quantizer_factory, layer_quantizer, for_reference)

      layer_data_type_map[layer] = LayerDataType(
          input_quantizer_list,
          multiplier,
          accumulator,
          None,
          None,
          None,
          None,
          output_quantizer,
          output_shapes,
          operation_count
      )

    # If it's a Quantized Activation layer.
    elif node_type in ["QActivation", "QAdaptiveActivation", "Activation"]:

      if for_reference or not hasattr(layer, "quantizer"):
        # Keras activation layer -> use default_interm_quantizer
        layer_quantizer = quantizer_factory.make_default_quantizer(
            mode=cfg.default_interm_quantizer)

        if keras_accumulator:
          layer_quantizer = quantizer_factory.make_default_quantizer(
              mode=keras_accumulator)
      else:
        layer_quantizer = layer.quantizer

        if not quantizer_factory.is_quantizer_supported(layer_quantizer):
          raise TagMissingError(
              "Unsupported activation quantizer {} on this layer: {}".format(
                  layer_quantizer, layer))

        if not layer_quantizer:
          layer_quantizer = quantizer_factory.make_default_quantizer(
              mode=cfg.default_interm_quantizer)

      output_quantizer = update_output_quantizer_in_graph(
          graph, node_id, quantizer_factory, layer_quantizer, for_reference)

      layer_data_type_map[layer] = LayerDataType(
          input_quantizer_list,
          None,
          None,
          None,
          w_shapes,
          None,
          b_shapes,
          output_quantizer,
          output_shapes,
          operation_count
      )

    elif node_type in ["QBatchNormalization", "BatchNormalization"]:

      (input_quantizer, _) = input_qe_list[0]

      # QKeras layers might be mixed with keras layers.
      if for_reference or not hasattr(layer, "get_quantizers"):
        # Keras BatchNorm layer mixed with quantized model
        # -> no reference mode
        gamma_quantizer = quantizer_factory.make_default_quantizer(
            mode=cfg.default_interm_quantizer)
        beta_quantizer = quantizer_factory.make_default_quantizer(
            mode=cfg.default_interm_quantizer)
        mean_quantizer = quantizer_factory.make_default_quantizer(
            mode=cfg.default_interm_quantizer)
        variance_quantizer = quantizer_factory.make_default_quantizer(
            mode=cfg.default_interm_quantizer)

        if keras_quantizer:
          gamma_quantizer = quantizer_factory.make_default_quantizer(
              mode=keras_quantizer)
          beta_quantizer = quantizer_factory.make_default_quantizer(
              mode=keras_quantizer)
          mean_quantizer = quantizer_factory.make_default_quantizer(
              mode=keras_quantizer)
          variance_quantizer = quantizer_factory.make_default_quantizer(
              mode=keras_quantizer)
      else:
        (qkeras_gamma_quantizer, qkeras_beta_quantizer, qkeras_mean_quantizer,
         qkeras_variance_quantizer) = layer.get_quantizers()

        if not qkeras_beta_quantizer:
          beta_quantizer = quantizer_factory.clone_quantizer(input_quantizer)
        else:
          beta_quantizer = quantizer_factory.make_quantizer(
              qkeras_beta_quantizer)

        if not qkeras_mean_quantizer:
          mean_quantizer = quantizer_factory.clone_quantizer(input_quantizer)
        else:
          mean_quantizer = quantizer_factory.make_quantizer(
              qkeras_mean_quantizer)

        if not qkeras_variance_quantizer:
          variance_quantizer = quantizer_factory.make_default_quantizer(
              mode=cfg.default_interm_quantizer)
        else:
          # If gamma is float, convert to input_quantizer.
          variance_quantizer = quantizer_factory.make_quantizer(
              qkeras_variance_quantizer)

        if not qkeras_gamma_quantizer:
          gamma_quantizer = quantizer_factory.make_default_quantizer(
              mode=cfg.default_interm_quantizer)
        else:
          gamma_quantizer = quantizer_factory.make_quantizer(
              qkeras_gamma_quantizer)

      # During inference, gamma, beta and variance are constants
      # if they are po2 quantizers, we need to modify their bits
      # with actual values and also update graph with the
      # corresponding output_quantizer on the edge.
      if is_inference:
        weights = qtools_util.get_weights(layer)
        # If no scale(gamma), num_weights --
        # If no center(beta_quantizer) num_weights --
        num_weights = 4
        if not layer.scale:
          num_weights -= 1
        if not layer.center:
          num_weights -= 1

        if layer.scale and gamma_quantizer.is_po2:
          gamma_quantizer.update_inference_values(weights[0])
        if variance_quantizer.is_po2:
          variance_quantizer.update_inference_values(
              weights[num_weights-1])

      qbn = quantized_operators.QBNFactory()
      qbn.make_quantizer(
          input_quantizer, gamma_quantizer, beta_quantizer,
          mean_quantizer, variance_quantizer, layer.scale, layer.center
      )

      def set_output(op, output):
        if op:
          op.output = output

      if for_reference or not hasattr(layer, "get_quantizers"):
        set_output(
            qbn.internal_divide_quantizer,
            quantizer_factory.make_default_quantizer(
                mode=cfg.default_interm_quantizer))

        set_output(
            qbn.internal_multiplier,
            quantizer_factory.make_default_quantizer(
                mode=cfg.default_interm_quantizer))

        set_output(
            qbn.internal_accumulator,
            quantizer_factory.make_default_quantizer(
                mode=cfg.default_interm_quantizer))

        set_output(
            qbn.internal_output,
            quantizer_factory.make_default_quantizer(
                mode=cfg.default_interm_quantizer))

        if keras_accumulator:
          set_output(
              qbn.internal_divide_quantizer,
              quantizer_factory.make_default_quantizer(mode=keras_accumulator))

          set_output(
              qbn.internal_multiplier,
              quantizer_factory.make_default_quantizer(mode=keras_accumulator))

          set_output(
              qbn.internal_accumulator,
              quantizer_factory.make_default_quantizer(mode=keras_accumulator))

          set_output(
              qbn.internal_output.output,
              quantizer_factory.make_default_quantizer(mode=keras_accumulator))

      layer_quantizer = qbn.internal_output.output

      output_quantizer = update_output_quantizer_in_graph(
          graph, node_id, quantizer_factory, layer_quantizer, for_reference)

      gamma_range = None
      if hasattr(layer, "gamma_range"):
        gamma_range = layer.gamma_range

      beta_range = None
      if hasattr(layer, "beta_range"):
        beta_range = layer.beta_range

      if not layer.center:
        qbn.beta_quantizer = None

      if not layer.scale:
        qbn.gamma_quantizer = None

      layer_data_type_map[layer] = {
          "input_quantizer_list": input_quantizer_list,
          "gamma_quantizer": gamma_quantizer,
          "beta_quantizer": beta_quantizer,
          "mean_quantizer": mean_quantizer,
          "variance_quantizer": variance_quantizer,
          "gamma_range": gamma_range,
          "beta_range": beta_range,
          "internal_divide_quantizer": qbn.internal_divide_quantizer,
          "internal_multiplier": qbn.internal_multiplier,
          "internal_accumulator": qbn.internal_accumulator,
          "output_quantizer": output_quantizer,
          "output_shapes": input_shape,
          "operation_count": operation_count
      }

    # If qdense, qconv, qpool, qoctave
    elif node_type in QKERAS_LAYERS or node_type in KERAS_LAYERS:

      (input_quantizer, _) = input_qe_list[0]

      if for_reference or not hasattr(layer, "get_quantizers"):
        # for_reference: force all quantizers to keras_quantizer
        weight_quantizer = quantizer_factory.make_default_quantizer(
            mode=cfg.default_interm_quantizer)
        bias_quantizer = quantizer_factory.make_default_quantizer(
            mode=cfg.default_interm_quantizer)

        if keras_quantizer:
          weight_quantizer = quantizer_factory.make_default_quantizer(
              mode=keras_quantizer)
          bias_quantizer = quantizer_factory.make_default_quantizer(
              mode=keras_quantizer)
      else:
        # qkeras layer
        qkeras_weight_quantizer = layer.get_quantizers()[0]
        qkeras_bias_quantizer = layer.get_quantizers()[1]

        if not quantizer_factory.is_quantizer_supported(
            qkeras_weight_quantizer):
          raise TagMissingError(
              "Unsupported weight quantizer {} on this layer: {}".format(
                  qkeras_weight_quantizer, layer))

        if not quantizer_factory.is_quantizer_supported(
            qkeras_bias_quantizer):
          raise TagMissingError(
              "Unsupported bias quantizer {} on this layer: {}".format(
                  qkeras_bias_quantizer, layer))

        weight_quantizer = quantizer_factory.make_quantizer(
            qkeras_weight_quantizer)
        bias_quantizer = quantizer_factory.make_quantizer(
            qkeras_bias_quantizer)

      # TODO(lishanok): During inference, if weight and bias is po2,
      #  need to update corresponding quantizer type with min and max
      #  of the constant values.
      if is_inference:
        weights = qtools_util.get_weights(layer)
        if weight_quantizer.is_po2:
          weight_quantizer.update_inference_values(weights[0])

        if bias_quantizer.is_po2:
          bias_quantizer.update_inference_values(weights[1])

      multiplier_factory = quantized_operators.MultiplierFactory()
      multiplier = multiplier_factory.make_multiplier(
          weight_quantizer, input_quantizer)

      weights = layer.get_weights()
      kernel = weights[0]

      accumulator_factory = quantized_operators.AccumulatorFactory()
      accumulator = accumulator_factory.make_accumulator(
          kernel.shape, multiplier)

      if not layer.use_bias:
        bias_quantizer = None

      if debug:
        print(layer.name or "None")
        print("weight_quantizer:", weight_quantizer.bits)
        print("input_quantizer:", input_quantizer.bits)
        print("multiplier_quantizer:", multiplier.output.bits)
        print("multiplier_gate_bits:", multiplier.gate_bits)
        print("accumulator:", accumulator.output.bits)

      if for_reference or not hasattr(layer, "get_quantizers"):
        accumulator.output = quantizer_factory.make_default_quantizer(
            mode=cfg.default_interm_quantizer)
        multiplier.output = quantizer_factory.make_default_quantizer(
            mode=cfg.default_interm_quantizer)

        if keras_accumulator:
          accumulator.output = quantizer_factory.make_default_quantizer(
              mode=keras_accumulator)
          multiplier.output = quantizer_factory.make_default_quantizer(
              mode=keras_accumulator)

      layer_quantizer = accumulator.output
      output_quantizer = update_output_quantizer_in_graph(
          graph, node_id, quantizer_factory, layer_quantizer, for_reference)

      layer_data_type_map[layer] = LayerDataType(
          input_quantizer_list,
          multiplier,
          accumulator,
          weight_quantizer,
          w_shapes,
          bias_quantizer,
          b_shapes,
          output_quantizer,
          output_shapes,
          operation_count
      )

    # Folded conv/dense/depthwiseconv layer
    elif node_type in ["QConv2DBatchnorm", "QDepthwiseConv2DBatchnorm"]:

      (input_quantizer, _) = input_qe_list[0]
      if for_reference or not hasattr(layer, "get_quantizers"):
        # For_reference: force all quantizers to keras_quantizer.
        weight_quantizer = quantizer_factory.make_default_quantizer(
            mode=cfg.default_interm_quantizer)
        bias_quantizer = quantizer_factory.make_default_quantizer(
            mode=cfg.default_interm_quantizer)

        if keras_quantizer:
          weight_quantizer = quantizer_factory.make_default_quantizer(
              mode=keras_quantizer)
          bias_quantizer = quantizer_factory.make_default_quantizer(
              mode=keras_quantizer)
      else:
        # QKeras layer
        qkeras_weight_quantizer = layer.get_quantizers()[0]
        qkeras_bias_quantizer = layer.get_quantizers()[1]
        if not quantizer_factory.is_quantizer_supported(
            qkeras_weight_quantizer):
          raise TagMissingError(
              "Unsupported weight quantizer {} on this layer: {}".format(
                  qkeras_weight_quantizer, layer))

        if not quantizer_factory.is_quantizer_supported(
            qkeras_bias_quantizer):
          raise TagMissingError(
              "Unsupported bias quantizer {} on this layer: {}".format(
                  qkeras_bias_quantizer, layer))

        weight_quantizer = quantizer_factory.make_quantizer(
            qkeras_weight_quantizer)

        if qkeras_bias_quantizer:
          bias_quantizer = quantizer_factory.make_quantizer(
              qkeras_bias_quantizer)
        else:
          bias_quantizer = None

      # TODO(lishanok): During inference, if weight and bias is po2,
      #  need to update corresponding quantizer type with min and max
      #  of the constant values
      if is_inference:
        weights = qtools_util.get_weights(layer)
        if weight_quantizer.is_po2:
          weight_quantizer.update_inference_values(weights[0])

        if bias_quantizer and bias_quantizer.is_po2:
          bias_quantizer.update_inference_values(weights[1])

      multiplier_factory = quantized_operators.MultiplierFactory()
      multiplier = multiplier_factory.make_multiplier(
          weight_quantizer, input_quantizer)

      weights = layer.get_weights()
      kernel = weights[0]

      accumulator_factory = quantized_operators.AccumulatorFactory()
      accumulator = accumulator_factory.make_accumulator(
          kernel.shape, multiplier, use_bias=True if bias_quantizer else False)

      if not bias_quantizer:
        # Sets bias the same as accumulator type.
        bias_quantizer = copy.deepcopy(accumulator.output)
        if not accumulator.output.is_floating_point:
          # For fixed point accumulator, needs to add 1 to its bits to avoid
          # possible satuation.
          accumulator.output.bits += 1
          accumulator.output.int_bits += 1
      if for_reference or not hasattr(layer, "get_quantizers"):
        accumulator.output = quantizer_factory.make_default_quantizer(
            mode=cfg.default_interm_quantizer)
        multiplier.output = quantizer_factory.make_default_quantizer(
            mode=cfg.default_interm_quantizer)

        if keras_accumulator:
          accumulator.output = quantizer_factory.make_default_quantizer(
              mode=keras_accumulator)
          multiplier.output = quantizer_factory.make_default_quantizer(
              mode=keras_accumulator)

      layer_quantizer = accumulator.output
      output_quantizer = update_output_quantizer_in_graph(
          graph, node_id, quantizer_factory, layer_quantizer, for_reference)

      layer_data_type_map[layer] = LayerDataType(
          input_quantizer_list,
          multiplier,
          accumulator,
          weight_quantizer,
          w_shapes,
          bias_quantizer,
          b_shapes,
          output_quantizer,
          output_shapes,
          operation_count
      )

    elif node_type:
      # Any other unsupported layer types -> pass the input quantizer
      # type to output in qraph
      (input_quantizer, _) = input_qe_list[0]

      if for_reference and keras_accumulator and not is_input_layer:
        input_quantizer = quantizer_factory.make_default_quantizer(
            mode=keras_accumulator)

      output_quantizer = update_output_quantizer_in_graph(
          graph, node_id, quantizer_factory, input_quantizer, for_reference)

      layer_data_type_map[layer] = LayerDataType(input_quantizer_list, None,
                                                 None, None, None, None, None,
                                                 output_quantizer,
                                                 output_shapes, operation_count)

  result = {
      "source_quantizer_list": source_quantizer_list,
      "output_layers": output_layers,
      "input_layers": input_layers,
      "layer_data_type_map": layer_data_type_map
  }

  return result



# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
""""create subtractor quantizer."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from qkeras.qtools.quantized_operators import adder_factory
from qkeras.qtools.quantized_operators import adder_impl
from qkeras.qtools.quantized_operators import quantizer_impl


class ISubtractor(adder_factory.IAdder):
  """Create a subtractor instance.

  The methods in subtractor is mostly inherited from adder
  with a few exceptions.
  """

  def make_quantizer(self, quantizer_1: quantizer_impl.IQuantizer,
                     quantizer_2: quantizer_impl.IQuantizer):
    """make an ISubtractor instance.

    if quantizer1 and quantizer2 are both non-signed, result should change
    to signed; else since a sign bit is already present,
    no need to add extra sign bit

    Args:
      quantizer_1: first operand
      quantizer_2: second operand

    Returns:
      An ISubtractor instance
    """
    quantizer = super().make_quantizer(quantizer_1, quantizer_2)

    if not isinstance(quantizer, adder_impl.FloatingPoint_Adder):
      if not quantizer_1.is_signed and not quantizer_2.is_signed:
        quantizer.output.is_signed = 1
        quantizer.output.bits += 1

    return quantizer



# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""adder operation implementation."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import abc

from qkeras.qtools.quantized_operators import accumulator_impl
from qkeras.qtools.quantized_operators import quantizer_impl


def po2_qbits_converter(po2_quantizer: quantizer_impl.IQuantizer):
  """convert a po2 quantizer to fixedpoint quantizer."""

  (bits_from_po2, int_bits_from_po2) = accumulator_impl.po2_to_qbits(
      po2_quantizer)
  qbits_quantizer = quantizer_impl.QuantizedBits()
  qbits_quantizer.bits = bits_from_po2
  qbits_quantizer.int_bits = int_bits_from_po2
  qbits_quantizer.is_signed = po2_quantizer.is_signed

  return qbits_quantizer


class IAdderImpl(abc.ABC):
  """abstract class for adder."""

  @staticmethod
  @abc.abstractmethod
  def implemented_as():
    pass


class FixedPointAdder(IAdderImpl):
  """adder for fixed point."""

  def __init__(self, quantizer_1, quantizer_2):
    self.output = quantizer_impl.QuantizedBits()
    self.output.bits = max(quantizer_1.bits, quantizer_2.bits) + 1
    self.output.int_bits = max(quantizer_1.int_bits,
                               quantizer_2.int_bits) + 1
    self.output.is_signed = quantizer_1.is_signed | quantizer_2.is_signed
    self.output.mode = 0
    self.output.is_floating_point = False
    self.output.is_po2 = 0

  @staticmethod
  def implemented_as():
    return "add"


class FloatingPointAdder(IAdderImpl):
  """floating point adder."""

  def __init__(self, quantizer_1, quantizer_2):
    bits = max(quantizer_1.bits, quantizer_2.bits)
    self.output = quantizer_impl.FloatingPoint(
        bits=bits)

  @staticmethod
  def implemented_as():
    return "add"


class Po2FixedPointAdder(IAdderImpl):
  """adder between po2 and fixed point."""

  def __init__(self, quantizer_1, quantizer_2):

    if quantizer_1.is_po2:
      po2_quantizer = quantizer_1
      fixedpoint_quantizer = quantizer_2
    else:
      po2_quantizer = quantizer_2
      fixedpoint_quantizer = quantizer_1

    # convert po2 to qbits first
    po2_qbits_quantizer = po2_qbits_converter(po2_quantizer)

    # qbits + qbits -> FixedPointAdder
    self.output = FixedPointAdder(po2_qbits_quantizer,
                                  fixedpoint_quantizer).output

  @staticmethod
  def implemented_as():
    return "add"


class Po2Adder(IAdderImpl):
  """adder for po2 type."""

  def __init__(self, quantizer_1, quantizer_2):
    qbits_quantizer_1 = po2_qbits_converter(quantizer_1)
    qbits_quantizer_2 = po2_qbits_converter(quantizer_2)
    self.output = FixedPointAdder(qbits_quantizer_1,
                                  qbits_quantizer_2).output

  @staticmethod
  def implemented_as():
    return "add"



# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""I/O implementation."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

from qkeras.qtools import generate_layer_data_type_map
from qkeras.qtools import qtools_util


def print_qstats(graph):
  """Prints quantization statistics for the model."""

  layer_data_type_map = generate_layer_data_type_map(graph)

  multipliers_counter = collections.Counter()

  print("")
  print("Number of operations in model:")
  for layer, data_type in layer_data_type_map.items():
    multiplier = data_type.multiplier
    multiplier_detail_str = "{}_{}, total_bits:{}, int_bits:{}".format(
        "signed" if multiplier.output.is_signed == 1 else "unsigned",
        multiplier.implemented_as(),
        multiplier.output.bits,
        multiplier.output.int_bits,
    )

    print("{}: {} x {}".format(
        layer.name,
        data_type.operation_count,
        multiplier_detail_str,
    ))

    multipliers_counter[
        multiplier_detail_str] += data_type.operation_count

  print("")
  print("Number of operation types in model:")
  for (multiplier_detail_str,
       total_multiplier_operation_count) in multipliers_counter.items():
    print("{}, x {}".format(multiplier_detail_str,
                            total_multiplier_operation_count))


def populate_quantizer(quantizer, shape=None, implemented_as=None):
  """write all the needed fields in the quantizer to dictionary."""

  mydict = collections.OrderedDict()

  if quantizer is not None:
    mydict["quantizer_type"] = quantizer.name

    # floats
    if quantizer.is_floating_point:
      mydict["bits"] = quantizer.bits

    # po2
    elif quantizer.is_po2:
      mydict["bits"] = quantizer.bits
      mydict["is_signed"] = quantizer.is_signed
      mydict["max_value"] = quantizer.max_val_po2

    # binary
    elif quantizer.mode in [3, 4]:

      mydict["bits"] = quantizer.bits
      mydict["int_bits"] = quantizer.int_bits
      mydict["is_signed"] = quantizer.is_signed
      if quantizer.mode == 4:
        mydict["values"] = [0, 1]
      else:
        mydict["values"] = [-1, 1]

    # ternary(-1, 0, 1)
    elif quantizer.mode == 2:
      mydict["bits"] = 2
      mydict["int_bits"] = 2
      mydict["is_signed"] = 1
      mydict["values"] = [0, -1, 1]

    # quantized_bits
    elif quantizer.mode == 0:
      mydict["bits"] = quantizer.bits
      mydict["int_bits"] = quantizer.int_bits + quantizer.is_signed
      mydict["is_signed"] = quantizer.is_signed

    if shape is not None:
      if isinstance(shape, tuple) and shape[0] is None:
        shape = list(shape)
        shape[0] = -1
        mydict["shape"] = tuple(shape)
      else:
        mydict["shape"] = shape

    if implemented_as is not None:
      mydict["op_type"] = implemented_as
  return mydict


def map_to_json(mydict):
  """write the dictionary to json format."""

  source_quantizer_list = mydict["source_quantizer_list"]
  layer_data_type_map = mydict["layer_data_type_map"]

  output_dict = collections.OrderedDict()

  q_list = []
  for source_quantizer in source_quantizer_list:
    tmp = populate_quantizer(source_quantizer)
    q_list.append(tmp)

  if bool(q_list):
    output_dict["source_quantizers"] = q_list

  for layer, feature in layer_data_type_map.items():
    layer_item = collections.OrderedDict()
    layer_item["layer_type"] = layer.__class__.__name__

    if layer_item["layer_type"] in [
        "QBatchNormalization", "BatchNormalization"]:
      layer_item["input_quantizer_list"] = [
          populate_quantizer(q) for q in feature["input_quantizer_list"]]

      if feature["gamma_quantizer"]:
        layer_item["gamma_quantizer"] = populate_quantizer(
            feature["gamma_quantizer"])

      if feature["beta_quantizer"]:
        layer_item["beta_quantizer"] = populate_quantizer(
            feature["beta_quantizer"])

      if feature["mean_quantizer"]:
        layer_item["mean_quantizer"] = populate_quantizer(
            feature["mean_quantizer"])

      if feature["variance_quantizer"]:
        layer_item["variance_quantizer"] = populate_quantizer(
            feature["variance_quantizer"])

      if feature["internal_divide_quantizer"]:
        layer_item["internal_divide_quantizer"] = populate_quantizer(
            feature["internal_divide_quantizer"].output,
            implemented_as=feature[
                "internal_divide_quantizer"].implemented_as())

      if feature["internal_multiplier"]:
        layer_item["internal_multiplier"] = populate_quantizer(
            feature["internal_multiplier"].output,
            implemented_as=feature[
                "internal_multiplier"].implemented_as())

      if feature["internal_accumulator"]:
        layer_item["internal_accumulator"] = populate_quantizer(
            feature["internal_accumulator"].output,
            implemented_as=feature["internal_accumulator"].implemented_as())

      if feature["output_quantizer"]:
        layer_item["output_quantizer"] = populate_quantizer(
            feature["output_quantizer"], shape=feature["output_shapes"])

    else:
      # populate the feature to dictionary
      layer_item["input_quantizer_list"] = [
          populate_quantizer(q) for q in feature.input_quantizer_list]

      tmp = populate_quantizer(feature.weight_quantizer, feature.w_shapes)
      if bool(tmp):
        layer_item["weight_quantizer"] = tmp

      tmp = populate_quantizer(feature.bias_quantizer, feature.b_shapes)
      if bool(tmp):
        layer_item["bias_quantizer"] = tmp

      if feature.multiplier:
        method = feature.multiplier.implemented_as()
        tmp = populate_quantizer(
            feature.multiplier.output,
            implemented_as=method)
        if bool(tmp):
          if qtools_util.is_merge_layers(layer):
            qname = layer.__class__.__name__ + "_quantizer"
            layer_item[qname] = tmp
          else:
            layer_item["multiplier"] = tmp

      if feature.accumulator:
        tmp = populate_quantizer(
            feature.accumulator.output,
            implemented_as=feature.accumulator.implemented_as())
        if bool(tmp):
          layer_item["accumulator"] = tmp

      tmp = populate_quantizer(feature.output_quantizer,
                               feature.output_shapes)
      if bool(tmp):
        layer_item["output_quantizer"] = tmp

      layer_item["operation_count"] = feature.operation_count

    output_dict[layer.name] = layer_item

  return output_dict



# Lint as: python3
# Copyright 2019 Google LLC
#
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""utility functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
import numpy as np
import tensorflow.keras.backend as K


def is_shape_alternation_layers(layer):
  lname = layer.__class__.__name__
  if lname:
    return "MaxPool" in lname or "Reshape" in lname or "Flatten" in lname
  return False


def is_merge_layers(layer):

  if layer.__class__.__name__ in [
      "Add", "Multiply", "Subtract", "Average", "Maximum", "Minimum",
      "Concatenate", "Dot"]:
    return True
  else:
    return False


def get_input_quantizers(graph, node_id, quantizer_factory, debug=False):
  """get the current layer's input quantizer."""

  # in merge layers, therea are more than 1 input

  output = []
  for parent_node_id in graph.predecessors(node_id):

    edge = graph.edges[(parent_node_id, node_id)]

    if debug:
      print("parent_node_id:", parent_node_id)
      print(edge)

    quantizer_on_edge = edge["quantizer"]
    input_quantizer = quantizer_factory.make_quantizer(quantizer_on_edge)

    output.append((input_quantizer, edge))

  return output


def get_input_quantizers_advanced(graph, node_id, is_input_layer, quantizer_factory,
                                  cfg, debug=False):
  """get input quantizer, deal with keras layer or lack of input quantizer in qkeras layer."""

  # in merge layers, therea are more than 1 input
  default_source_quantizer = cfg.default_source_quantizer
  default_interm_quantizer = cfg.default_interm_quantizer

  output = []
  for parent_node_id in graph.predecessors(node_id):

    edge = graph.edges[(parent_node_id, node_id)]

    if debug:
      print("parent_node_id:", parent_node_id)
      print(edge)

    quantizer_on_edge = edge["quantizer"]
    input_quantizer = quantizer_factory.make_quantizer(quantizer_on_edge)

    if is_input_layer and not input_quantizer:
      # input layer without input_quantizer specified->use default_source_quantizer
      input_quantizer = quantizer_factory.make_default_quantizer(
          mode=default_source_quantizer)
    elif not input_quantizer:
      # if no input quantizer is available-> use default quantizer from config.json
      input_quantizer = quantizer_factory.make_default_quantizer(
          mode=default_interm_quantizer)

    output.append((input_quantizer, edge))

  return output


def get_operation_count(layer, input_shape):
  """Determines number of multiplier operations in a qkeras layer."""

  # Check if the inputs are a list of Dimensions
  if isinstance(input_shape, list):
    input_shape = input_shape[0]

  operation_count = 0

  if is_merge_layers(layer) or is_shape_alternation_layers(layer):
    operation_count = np.prod(input_shape[1:])

  elif layer.__class__.__name__ in [
      "AveragePooling2D", "AvgPool2D", "GlobalAvgPool2D",
      "GlobalAveragePooling2D"
  ]:

    if hasattr(layer, "pool_size"):
      pool_size = layer.pool_size
    else:
      pool_size = input_shape[1:-1]
    add_ops = np.prod(pool_size)

    output_shape = layer.compute_output_shape(input_shape)
    channels_o = output_shape[-1]

    # total number of add ops
    operation_count = channels_o * add_ops

  elif "UpSampling" in layer.__class__.__name__:
    # UpSampling1D/2D/3D
    output_shape = layer.compute_output_shape(input_shape)
    operation_count = np.prod(output_shape[1:])

  elif ("Activation" in layer.__class__.__name__ or
        "BatchNormalization" in layer.__class__.__name__):
    operation_count = np.prod(input_shape[1:])

  elif layer.__class__.__name__ in ["QConv2D", "Conv2D", "QConv2DBatchnorm"]:

    output_shape = layer.compute_output_shape(input_shape)
    _, _, _, channels_i = input_shape

    _, height_o, width_o, channels_o = output_shape

    weight = layer.get_weights()[0]

    kernel_h, kernel_w, _, _ = weight.shape

    operation_count = (
        height_o * width_o * channels_o * kernel_h * kernel_w * channels_i)

  elif layer.__class__.__name__ in ["QConv1D", "Conv1D"]:
    output_shape = layer.compute_output_shape(input_shape)
    _, _, channels_i = input_shape

    _, time_o, channels_o = output_shape

    weight = layer.get_weights()[0]

    kernel_length, _, _ = weight.shape

    operation_count = (
        time_o * channels_o * kernel_length * channels_i)

  elif layer.__class__.__name__ in ["QDepthwiseConv2D", "DepthwiseConv2D"]:
    output_shape = layer.compute_output_shape(input_shape)
    _, _, _, channels_i = input_shape

    _, height_o, width_o, channels_o = output_shape

    weight_1 = layer.get_weights()[0]

    kernel_h, kernel_w, _, _ = weight_1.shape

    operation_count = (
        kernel_h * kernel_w * height_o * width_o * channels_i)

  elif layer.__class__.__name__ in ["QDense", "Dense"]:
    output_shape = layer.compute_output_shape(input_shape)
    _, size_i = input_shape
    _, size_o = output_shape

    operation_count = (size_i * size_o)

  else:
    print("operation count for {} is defaulted to 0".format(
        layer))

  return int(operation_count)


def get_weights(layer):
  weights = layer.get_weights()
  out = copy.deepcopy(weights)
  for j, weight in enumerate(weights):
    if hasattr(layer, "get_quantizers") and layer.get_quantizers()[j]:
      out[j] = K.eval(
          layer.get_quantizers()[j](K.constant(weight)))

  return out
